from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1709() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.OntologyAnnotation.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("OntologyAnnotation", string_type), ("annotation_value", string_type), ("term_source", string_type), ("term_accession", string_type), ("comments", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    OntologyAnnotation: str
    annotation_value: str
    term_source: str
    term_accession: str
    comments: str

IContext_reflection = _expr1709

def _arrow1717(__unit: None=None) -> IEncodable:
    class ObjectExpr1710(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1711(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("sdo:DefinedTerm")

    class ObjectExpr1712(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:name")

    class ObjectExpr1713(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:inDefinedTermSet")

    class ObjectExpr1714(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("sdo:termCode")

    class ObjectExpr1715(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            return helpers_5.encode_string("sdo:disambiguatingDescription")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1710()), ("OntologyAnnotation", ObjectExpr1711()), ("annotationValue", ObjectExpr1712()), ("termSource", ObjectExpr1713()), ("termAccession", ObjectExpr1714()), ("comments", ObjectExpr1715())])
    class ObjectExpr1716(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_6))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_6.encode_object(arg)

    return ObjectExpr1716()


context_jsonvalue: IEncodable = _arrow1717()

__all__ = ["IContext_reflection", "context_jsonvalue"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from ...fable_library.option import (default_arg, map, bind)
from ...fable_library.reflection import (TypeInfo, option_type, record_type)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (to_string, Record)
from ...fable_library.util import int32_to_string
from ..ontology_annotation import (OntologyAnnotation, OntologyAnnotation_reflection)
from ..value import (Value as Value_1, Value_reflection)
from .protocol_parameter import (ProtocolParameter, ProtocolParameter_reflection)

def _expr773() -> TypeInfo:
    return record_type("ARCtrl.Process.ProcessParameterValue", [], ProcessParameterValue, lambda: [("Category", option_type(ProtocolParameter_reflection())), ("Value", option_type(Value_reflection())), ("Unit", option_type(OntologyAnnotation_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class ProcessParameterValue(Record):
    Category: ProtocolParameter | None
    Value: Value_1 | None
    Unit: OntologyAnnotation | None
    @staticmethod
    def make(category: ProtocolParameter | None=None, value: Value_1 | None=None, unit: OntologyAnnotation | None=None) -> ProcessParameterValue:
        return ProcessParameterValue(category, value, unit)

    @staticmethod
    def create(Category: ProtocolParameter | None=None, Value: Value_1 | None=None, Unit: OntologyAnnotation | None=None) -> ProcessParameterValue:
        return ProcessParameterValue.make(Category, Value, Unit)

    @staticmethod
    def empty() -> ProcessParameterValue:
        return ProcessParameterValue.create()

    @property
    def NameText(self, __unit: None=None) -> str:
        this: ProcessParameterValue = self
        def mapping(oa: ProtocolParameter) -> str:
            return oa.NameText

        return default_arg(map(mapping, this.Category), "")

    @property
    def TryNameText(self, __unit: None=None) -> str | None:
        this: ProcessParameterValue = self
        def binder(oa: ProtocolParameter) -> str | None:
            return oa.TryNameText

        return bind(binder, this.Category)

    @property
    def ValueText(self, __unit: None=None) -> str:
        this: ProcessParameterValue = self
        def mapping(oa: Value_1) -> str:
            if oa.tag == 2:
                return to_string(oa.fields[0])

            elif oa.tag == 1:
                return int32_to_string(oa.fields[0])

            elif oa.tag == 3:
                return oa.fields[0]

            else: 
                return oa.fields[0].NameText


        return default_arg(map(mapping, this.Value), "")

    @property
    def ValueWithUnitText(self, __unit: None=None) -> str:
        this: ProcessParameterValue = self
        def mapping(oa: OntologyAnnotation) -> str:
            return oa.NameText

        unit: str | None = map(mapping, this.Unit)
        v: str = this.ValueText
        if unit is None:
            return v

        else: 
            u: str = unit
            return to_text(printf("%s %s"))(v)(u)


    def MapCategory(self, f: Callable[[OntologyAnnotation], OntologyAnnotation]) -> ProcessParameterValue:
        this: ProcessParameterValue = self
        def mapping(p: ProtocolParameter) -> ProtocolParameter:
            return p.MapCategory(f)

        return ProcessParameterValue(map(mapping, this.Category), this.Value, this.Unit)

    def SetCategory(self, c: OntologyAnnotation) -> ProcessParameterValue:
        this: ProcessParameterValue = self
        def _arrow769(__unit: None=None) -> ProtocolParameter | None:
            match_value: ProtocolParameter | None = this.Category
            if match_value is None:
                return ProtocolParameter.create(None, c)

            else: 
                p: ProtocolParameter = match_value
                return p.SetCategory(c)


        return ProcessParameterValue(_arrow769(), this.Value, this.Unit)

    @staticmethod
    def try_get_name_text(pv: ProcessParameterValue) -> str | None:
        return pv.TryNameText

    @staticmethod
    def get_name_text(pv: ProcessParameterValue) -> str:
        return pv.NameText

    @staticmethod
    def name_equals_string(name: str, pv: ProcessParameterValue) -> bool:
        return pv.NameText == name

    @staticmethod
    def get_category(pv: ProcessParameterValue) -> ProtocolParameter | None:
        return pv.Category

    @staticmethod
    def create_as_pv(alternate_name: str | None=None, measurement_method: str | None=None, description: str | None=None, category: OntologyAnnotation | None=None, value: Value_1 | None=None, unit: OntologyAnnotation | None=None) -> ProcessParameterValue:
        def mapping(c: OntologyAnnotation) -> ProtocolParameter:
            return ProtocolParameter.create(None, c)

        category_1: ProtocolParameter | None = map(mapping, category)
        return ProcessParameterValue.create(category_1, value, unit)

    def Print(self, __unit: None=None) -> str:
        this: ProcessParameterValue = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: ProcessParameterValue = self
        def mapping(f: ProtocolParameter) -> str:
            return f.NameText

        category: str | None = map(mapping, this.Category)
        def mapping_1(oa: OntologyAnnotation) -> str:
            return oa.NameText

        unit: str | None = map(mapping_1, this.Unit)
        def mapping_2(v: Value_1) -> str:
            s: str = v.PrintCompact()
            if unit is None:
                return s

            else: 
                return (s + " ") + unit


        value: str | None = map(mapping_2, this.Value)
        def _arrow770(__unit: None=None) -> str:
            value_2: str = value
            return value_2

        def _arrow771(__unit: None=None) -> str:
            category_2: str = category
            return (category_2 + ":") + "No Value"

        def _arrow772(__unit: None=None) -> str:
            category_1: str = category
            value_1: str = value
            return (category_1 + ":") + value_1

        return ("" if (value is None) else _arrow770()) if (category is None) else (_arrow771() if (value is None) else _arrow772())

    def AlternateName(self, __unit: None=None) -> str | None:
        return None

    def MeasurementMethod(self, __unit: None=None) -> str | None:
        return None

    def Description(self, __unit: None=None) -> str | None:
        return None

    def GetCategory(self, __unit: None=None) -> OntologyAnnotation | None:
        this: ProcessParameterValue = self
        def binder(p: ProtocolParameter) -> OntologyAnnotation | None:
            return p.ParameterName

        return bind(binder, this.Category)

    def GetAdditionalType(self, __unit: None=None) -> str:
        return "ProcessParameterValue"

    def GetValue(self, __unit: None=None) -> Value_1 | None:
        this: ProcessParameterValue = self
        return this.Value

    def GetUnit(self, __unit: None=None) -> OntologyAnnotation | None:
        this: ProcessParameterValue = self
        return this.Unit


ProcessParameterValue_reflection = _expr773

__all__ = ["ProcessParameterValue_reflection"]




from __future__ import annotations
import json as json_2
from typing import Any
from ..fable_library.list import (to_array, FSharpList)
from ..fable_library.seq import to_array as to_array_1
from ..fable_library.types import (Array, uint32)
from ..fable_library.util import (get_enumerator, IEnumerable_1)
from ..thoth_json_core.types import (IEncoderHelpers_1, IEncodable)

class ObjectExpr389(IEncoderHelpers_1[Any]):
    def encode_string(self, value: str) -> Any:
        return value

    def encode_char(self, value_1: str) -> Any:
        return value_1

    def encode_decimal_number(self, value_2: float) -> Any:
        return value_2

    def encode_bool(self, value_3: bool) -> Any:
        return value_3

    def encode_null(self, __unit: None=None) -> Any:
        return None

    def encode_object(self, values: IEnumerable_1[tuple[str, Any]]) -> Any:
        o: Any = {}
        with get_enumerator(values) as enumerator:
            while enumerator.System_Collections_IEnumerator_MoveNext():
                for_loop_var: tuple[str, Any] = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                o[for_loop_var[0]] = for_loop_var[1]
        return o

    def encode_array(self, values_1: Array[Any]) -> Any:
        return values_1

    def encode_list(self, values_2: FSharpList[Any]) -> Any:
        this: IEncoderHelpers_1[Any] = self
        arg: Array[Any] = to_array(values_2)
        return this.encode_array(arg)

    def encode_seq(self, values_3: IEnumerable_1[Any]) -> Any:
        this_1: IEncoderHelpers_1[Any] = self
        arg_1: Array[Any] = to_array_1(values_3)
        return this_1.encode_array(arg_1)

    def encode_resize_array(self, values_4: Array[Any]) -> Any:
        this_2: IEncoderHelpers_1[Any] = self
        arg_2: Array[Any] = values_4[:]
        return this_2.encode_array(arg_2)

    def encode_signed_integral_number(self, value_5: int) -> Any:
        return value_5

    def encode_unsigned_integral_number(self, value_6: uint32) -> Any:
        return value_6


helpers: IEncoderHelpers_1[Any] = ObjectExpr389()

def to_string(space: int, value: IEncodable) -> str:
    json_1: Any = value.Encode(helpers)
    if space == 0:
        return json_2.dumps(json_1, separators = [",", ":"], ensure_ascii = False)

    else: 
        return json_2.dumps(json_1, indent = space, ensure_ascii = False)



__all__ = ["helpers", "to_string"]




from __future__ import annotations
from ..arctrl_core.arc_types import (ArcAssay, ArcStudy, ArcWorkflow, ArcRun, ArcInvestigation)
from ..arctrl_core.data_map import DataMap
from ..arctrl_spreadsheet.arc_assay import (ARCtrl_ArcAssay__ArcAssay_fromFsWorkbook_Static_32154C9D, ARCtrl_ArcAssay__ArcAssay_toFsWorkbook_Static_Z2508BE4F)
from ..arctrl_spreadsheet.arc_investigation import (ARCtrl_ArcInvestigation__ArcInvestigation_fromFsWorkbook_Static_32154C9D, ARCtrl_ArcInvestigation__ArcInvestigation_toFsWorkbook_Static_Z720BD3FF)
from ..arctrl_spreadsheet.arc_run import (ARCtrl_ArcRun__ArcRun_fromFsWorkbook_Static_32154C9D, ARCtrl_ArcRun__ArcRun_toFsWorkbook_Static_Z3EFAF6F8)
from ..arctrl_spreadsheet.arc_study import (ARCtrl_ArcStudy__ArcStudy_fromFsWorkbook_Static_32154C9D, ARCtrl_ArcStudy__ArcStudy_toFsWorkbook_Static_Z4CEFA522)
from ..arctrl_spreadsheet.arc_workflow import (ARCtrl_ArcWorkflow__ArcWorkflow_fromFsWorkbook_Static_32154C9D, ARCtrl_ArcWorkflow__ArcWorkflow_toFsWorkbook_Static_Z1C75CB0E)
from ..arctrl_spreadsheet.data_map import (from_fs_workbook, to_fs_workbook)
from ..fable_library.list import (FSharpList, of_seq)
from ..fable_library.option import map
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.types import Array
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from ..fs_spreadsheet_py.fs_extension import (FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxFile_Static_Z721C83C5, FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static)

def _expr4093() -> TypeInfo:
    return class_type("ARCtrl.XlsxHelper.DatamapXlsx", None, XlsxHelper_DatamapXlsx)


class XlsxHelper_DatamapXlsx:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_fs_workbook(self, fswb: FsWorkbook) -> DataMap:
        return from_fs_workbook(fswb)

    def to_fs_workbook(self, datamap: DataMap) -> FsWorkbook:
        return to_fs_workbook(datamap)

    def from_xlsx_file(self, path: str) -> DataMap:
        return from_fs_workbook(FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxFile_Static_Z721C83C5(path))

    def to_xlsx_file(self, path: str, datamap: DataMap) -> None:
        FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static(path, to_fs_workbook(datamap))


XlsxHelper_DatamapXlsx_reflection = _expr4093

def XlsxHelper_DatamapXlsx__ctor(__unit: None=None) -> XlsxHelper_DatamapXlsx:
    return XlsxHelper_DatamapXlsx(__unit)


def _expr4094() -> TypeInfo:
    return class_type("ARCtrl.XlsxHelper.AssayXlsx", None, XlsxHelper_AssayXlsx)


class XlsxHelper_AssayXlsx:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_fs_workbook(self, fswb: FsWorkbook) -> ArcAssay:
        return ARCtrl_ArcAssay__ArcAssay_fromFsWorkbook_Static_32154C9D(fswb)

    def to_fs_workbook(self, assay: ArcAssay) -> FsWorkbook:
        return ARCtrl_ArcAssay__ArcAssay_toFsWorkbook_Static_Z2508BE4F(assay)

    def from_xlsx_file(self, path: str) -> ArcAssay:
        return ARCtrl_ArcAssay__ArcAssay_fromFsWorkbook_Static_32154C9D(FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxFile_Static_Z721C83C5(path))

    def to_xlsx_file(self, path: str, assay: ArcAssay) -> None:
        FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static(path, ARCtrl_ArcAssay__ArcAssay_toFsWorkbook_Static_Z2508BE4F(assay))


XlsxHelper_AssayXlsx_reflection = _expr4094

def XlsxHelper_AssayXlsx__ctor(__unit: None=None) -> XlsxHelper_AssayXlsx:
    return XlsxHelper_AssayXlsx(__unit)


def _expr4095() -> TypeInfo:
    return class_type("ARCtrl.XlsxHelper.StudyXlsx", None, XlsxHelper_StudyXlsx)


class XlsxHelper_StudyXlsx:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_fs_workbook(self, fswb: FsWorkbook) -> tuple[ArcStudy, FSharpList[ArcAssay]]:
        return ARCtrl_ArcStudy__ArcStudy_fromFsWorkbook_Static_32154C9D(fswb)

    def to_fs_workbook(self, study: ArcStudy, assays: Array[ArcAssay] | None=None) -> FsWorkbook:
        return ARCtrl_ArcStudy__ArcStudy_toFsWorkbook_Static_Z4CEFA522(study, map(of_seq, assays))

    def from_xlsx_file(self, path: str) -> tuple[ArcStudy, FSharpList[ArcAssay]]:
        return ARCtrl_ArcStudy__ArcStudy_fromFsWorkbook_Static_32154C9D(FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxFile_Static_Z721C83C5(path))

    def to_xlsx_file(self, path: str, study: ArcStudy, assays: Array[ArcAssay] | None=None) -> None:
        FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static(path, ARCtrl_ArcStudy__ArcStudy_toFsWorkbook_Static_Z4CEFA522(study, map(of_seq, assays)))


XlsxHelper_StudyXlsx_reflection = _expr4095

def XlsxHelper_StudyXlsx__ctor(__unit: None=None) -> XlsxHelper_StudyXlsx:
    return XlsxHelper_StudyXlsx(__unit)


def _expr4096() -> TypeInfo:
    return class_type("ARCtrl.XlsxHelper.WorkflowXlsx", None, XlsxHelper_WorkflowXlsx)


class XlsxHelper_WorkflowXlsx:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_fs_workbook(self, fswb: FsWorkbook) -> ArcWorkflow:
        return ARCtrl_ArcWorkflow__ArcWorkflow_fromFsWorkbook_Static_32154C9D(fswb)

    def to_fs_workbook(self, workflow: ArcWorkflow) -> FsWorkbook:
        return ARCtrl_ArcWorkflow__ArcWorkflow_toFsWorkbook_Static_Z1C75CB0E(workflow)

    def from_xlsx_file(self, path: str) -> ArcWorkflow:
        return ARCtrl_ArcWorkflow__ArcWorkflow_fromFsWorkbook_Static_32154C9D(FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxFile_Static_Z721C83C5(path))

    def to_xlsx_file(self, path: str, workflow: ArcWorkflow) -> None:
        FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static(path, ARCtrl_ArcWorkflow__ArcWorkflow_toFsWorkbook_Static_Z1C75CB0E(workflow))


XlsxHelper_WorkflowXlsx_reflection = _expr4096

def XlsxHelper_WorkflowXlsx__ctor(__unit: None=None) -> XlsxHelper_WorkflowXlsx:
    return XlsxHelper_WorkflowXlsx(__unit)


def _expr4097() -> TypeInfo:
    return class_type("ARCtrl.XlsxHelper.RunXlsx", None, XlsxHelper_RunXlsx)


class XlsxHelper_RunXlsx:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_fs_workbook(self, fswb: FsWorkbook) -> ArcRun:
        return ARCtrl_ArcRun__ArcRun_fromFsWorkbook_Static_32154C9D(fswb)

    def to_fs_workbook(self, run: ArcRun) -> FsWorkbook:
        return ARCtrl_ArcRun__ArcRun_toFsWorkbook_Static_Z3EFAF6F8(run)

    def from_xlsx_file(self, path: str) -> ArcRun:
        return ARCtrl_ArcRun__ArcRun_fromFsWorkbook_Static_32154C9D(FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxFile_Static_Z721C83C5(path))

    def to_xlsx_file(self, path: str, run: ArcRun) -> None:
        FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static(path, ARCtrl_ArcRun__ArcRun_toFsWorkbook_Static_Z3EFAF6F8(run))


XlsxHelper_RunXlsx_reflection = _expr4097

def XlsxHelper_RunXlsx__ctor(__unit: None=None) -> XlsxHelper_RunXlsx:
    return XlsxHelper_RunXlsx(__unit)


def _expr4098() -> TypeInfo:
    return class_type("ARCtrl.XlsxHelper.InvestigationXlsx", None, XlsxHelper_InvestigationXlsx)


class XlsxHelper_InvestigationXlsx:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_fs_workbook(self, fswb: FsWorkbook) -> ArcInvestigation:
        return ARCtrl_ArcInvestigation__ArcInvestigation_fromFsWorkbook_Static_32154C9D(fswb)

    def to_fs_workbook(self, investigation: ArcInvestigation) -> FsWorkbook:
        return ARCtrl_ArcInvestigation__ArcInvestigation_toFsWorkbook_Static_Z720BD3FF(investigation)

    def from_xlsx_file(self, path: str) -> ArcInvestigation:
        return ARCtrl_ArcInvestigation__ArcInvestigation_fromFsWorkbook_Static_32154C9D(FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxFile_Static_Z721C83C5(path))

    def to_xlsx_file(self, path: str, investigation: ArcInvestigation) -> None:
        FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static(path, ARCtrl_ArcInvestigation__ArcInvestigation_toFsWorkbook_Static_Z720BD3FF(investigation))


XlsxHelper_InvestigationXlsx_reflection = _expr4098

def XlsxHelper_InvestigationXlsx__ctor(__unit: None=None) -> XlsxHelper_InvestigationXlsx:
    return XlsxHelper_InvestigationXlsx(__unit)


def _expr4099() -> TypeInfo:
    return class_type("ARCtrl.XlsxController", None, XlsxController)


class XlsxController:
    @staticmethod
    def Datamap() -> XlsxHelper_DatamapXlsx:
        return XlsxHelper_DatamapXlsx()

    @staticmethod
    def Assay() -> XlsxHelper_AssayXlsx:
        return XlsxHelper_AssayXlsx()

    @staticmethod
    def Study() -> XlsxHelper_StudyXlsx:
        return XlsxHelper_StudyXlsx()

    @staticmethod
    def Workflow() -> XlsxHelper_WorkflowXlsx:
        return XlsxHelper_WorkflowXlsx()

    @staticmethod
    def Run() -> XlsxHelper_RunXlsx:
        return XlsxHelper_RunXlsx()

    @staticmethod
    def Investigation() -> XlsxHelper_InvestigationXlsx:
        return XlsxHelper_InvestigationXlsx()


XlsxController_reflection = _expr4099

__all__ = ["XlsxHelper_DatamapXlsx_reflection", "XlsxHelper_AssayXlsx_reflection", "XlsxHelper_StudyXlsx_reflection", "XlsxHelper_WorkflowXlsx_reflection", "XlsxHelper_RunXlsx_reflection", "XlsxHelper_InvestigationXlsx_reflection", "XlsxController_reflection"]




from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, TypeVar

from .array_ import copy_to, fill, initialize
from .double import divide
from .reflection import TypeInfo, class_type
from .seq import append, delay, empty, enumerate_while, singleton, to_array
from .types import Array, FSharpRef
from .util import (
    IEnumerable_1,
    IEnumerator,
    compare,
    compare_primitives,
    get_enumerator,
    max,
    structural_hash,
    to_iterator,
)
from .util import equals as equals_1


_T = TypeVar("_T")


def _expr4(gen0: TypeInfo) -> TypeInfo:
    return class_type("System.Collections.Generic.Comparer`1", [gen0], Comparer_1)


class Comparer_1(Generic[_T]):
    def __init__(self, comparison: Callable[[_T, _T], int]) -> None:
        self.comparison: Callable[[_T, _T], int] = comparison

    def Compare(self, x: _T, y: _T) -> int:
        _: Comparer_1[_T] = self
        return _.comparison(x, y)


Comparer_1_reflection = _expr4


def Comparer_1__ctor_47C913C(comparison: Callable[[_T, _T], int]) -> Comparer_1[_T]:
    return Comparer_1(comparison)


def Comparer_1_get_Default(__unit: None = None) -> Comparer_1[Any]:
    return Comparer_1__ctor_47C913C(compare)


def Comparer_1_Create_47C913C(comparison: Callable[[_T, _T], int]) -> Comparer_1[Any]:
    return Comparer_1__ctor_47C913C(comparison)


def Comparer_1__Compare_5BDDA0(_: Comparer_1[Any], x: Any, y: Any) -> int:
    return _.comparison(x, y)


def _expr7(gen0: TypeInfo) -> TypeInfo:
    return class_type("System.Collections.Generic.EqualityComparer`1", [gen0], EqualityComparer_1)


class EqualityComparer_1(Generic[_T]):
    def __init__(self, equals: Callable[[_T, _T], bool], get_hash_code: Callable[[_T], int]) -> None:
        self.equals: Callable[[_T, _T], bool] = equals
        self.get_hash_code: Callable[[_T], int] = get_hash_code

    def __eq__(self, x: _T, y: _T) -> bool:
        _: EqualityComparer_1[_T] = self
        return _.equals(x, y)

    def __hash__(self, x: _T | None = None) -> int:
        _: EqualityComparer_1[_T] = self
        return _.get_hash_code(x)


EqualityComparer_1_reflection = _expr7


def EqualityComparer_1__ctor_Z6EE254AB(
    equals: Callable[[_T, _T], bool], get_hash_code: Callable[[_T], int]
) -> EqualityComparer_1[_T]:
    return EqualityComparer_1(equals, get_hash_code)


def EqualityComparer_1_get_Default(__unit: None = None) -> EqualityComparer_1[Any]:
    return EqualityComparer_1__ctor_Z6EE254AB(equals_1, structural_hash)


def EqualityComparer_1_Create_Z6EE254AB(
    equals: Callable[[_T, _T], bool], get_hash_code: Callable[[_T], int]
) -> EqualityComparer_1[Any]:
    return EqualityComparer_1__ctor_Z6EE254AB(equals, get_hash_code)


def EqualityComparer_1__Equals_5BDDA0(_: EqualityComparer_1[Any], x: Any, y: Any) -> bool:
    return _.equals(x, y)


def EqualityComparer_1__GetHashCode_2B595(_: EqualityComparer_1[Any], x: Any) -> int:
    return _.get_hash_code(x)


def _expr12(gen0: TypeInfo) -> TypeInfo:
    return class_type("System.Collections.Generic.Stack`1", [gen0], Stack_1)


class Stack_1(Generic[_T]):
    def __init__(self, initial_contents: Array[Any], initial_count: int) -> None:
        self.contents: Array[_T] = initial_contents
        self.count: int = initial_count or 0

    def GetEnumerator(self, __unit: None = None) -> IEnumerator[_T]:
        _: Stack_1[_T] = self

        def _arrow11(__unit: None = None) -> IEnumerable_1[_T]:
            index: int = (_.count - 1) or 0

            def _arrow8(__unit: None = None) -> bool:
                return index >= 0

            def _arrow10(__unit: None = None) -> IEnumerable_1[_T]:
                def _arrow9(__unit: None = None) -> IEnumerable_1[_T]:
                    nonlocal index
                    index = (index - 1) or 0
                    return empty()

                return append(singleton(_.contents[index]), delay(_arrow9))

            return enumerate_while(_arrow8, delay(_arrow10))

        return get_enumerator(delay(_arrow11))

    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None = None) -> IEnumerator[Any]:
        this: Stack_1[_T] = self
        return get_enumerator(this)


Stack_1_reflection = _expr12


def Stack_1__ctor_Z3B4C077E(initial_contents: Array[Any], initial_count: int) -> Stack_1[_T]:
    return Stack_1(initial_contents, initial_count)


def Stack_1__ctor_Z524259A4(initial_capacity: int) -> Stack_1[Any]:
    return Stack_1__ctor_Z3B4C077E(fill([0] * initial_capacity, 0, initial_capacity, None), 0)


def Stack_1__ctor(__unit: None = None) -> Stack_1[Any]:
    return Stack_1__ctor_Z524259A4(4)


def Stack_1__ctor_BB573A(xs: IEnumerable_1[Any]) -> Stack_1[Any]:
    arr: Array[_T] = list(xs)
    return Stack_1__ctor_Z3B4C077E(arr, len(arr))


def Stack_1__Ensure_Z524259A4(_: Stack_1[Any], new_size: int) -> None:
    old_size: int = len(_.contents) or 0
    if new_size > old_size:
        old: Array[_T] = _.contents

        def _arrow14(x: int, y: int, _: Any = _, new_size: Any = new_size) -> int:
            return compare_primitives(x, y)

        def _arrow15(x: int, y: int, _: Any = _, new_size: Any = new_size) -> int:
            return compare_primitives(x, y)

        _.contents = fill([0] * max(_arrow14, new_size, old_size * 2), 0, max(_arrow15, new_size, old_size * 2), None)
        copy_to(old, 0, _.contents, 0, _.count)


def Stack_1__get_Count(_: Stack_1[Any]) -> int:
    return _.count


def Stack_1__Pop(_: Stack_1[Any]) -> Any:
    _.count = (_.count - 1) or 0
    return _.contents[_.count]


def Stack_1__Peek(_: Stack_1[Any]) -> Any:
    return _.contents[_.count - 1]


def Stack_1__Contains_2B595(_: Stack_1[Any], x: Any) -> bool:
    found: bool = False
    i: int = 0
    while (not found) if (i < _.count) else False:
        if equals_1(x, _.contents[i]):
            found = True

        else:
            i = (i + 1) or 0

    return found


def Stack_1__TryPeek_1F3DB691(this: Stack_1[Any], result: FSharpRef[_T]) -> bool:
    if this.count > 0:
        result.contents = Stack_1__Peek(this)
        return True

    else:
        return False


def Stack_1__TryPop_1F3DB691(this: Stack_1[Any], result: FSharpRef[_T]) -> bool:
    if this.count > 0:
        result.contents = Stack_1__Pop(this)
        return True

    else:
        return False


def Stack_1__Push_2B595(this: Stack_1[Any], x: Any) -> None:
    Stack_1__Ensure_Z524259A4(this, this.count + 1)
    this.contents[this.count] = x
    this.count = (this.count + 1) or 0


def Stack_1__Clear(_: Stack_1[Any]) -> None:
    _.count = 0
    fill(_.contents, 0, len(_.contents), None)


def Stack_1__TrimExcess(this: Stack_1[Any]) -> None:
    if divide(this.count, len(this.contents)) > 0.9:
        Stack_1__Ensure_Z524259A4(this, this.count)


def Stack_1__ToArray(_: Stack_1[Any]) -> Array[Any]:
    def _arrow16(i: int, _: Any = _) -> _T:
        return _.contents[(_.count - 1) - i]

    return initialize(_.count, _arrow16, None)


def _expr17(gen0: TypeInfo) -> TypeInfo:
    return class_type("System.Collections.Generic.Queue`1", [gen0], Queue_1)


class Queue_1(Generic[_T]):
    def __init__(self, initial_contents: Array[Any], initial_count: int) -> None:
        self.contents: Array[_T] = initial_contents
        self.count: int = initial_count or 0
        self.head: int = 0
        self.tail: int = (0 if (initial_count == len(self.contents)) else initial_count) or 0

    def GetEnumerator(self, __unit: None = None) -> IEnumerator[_T]:
        _: Queue_1[_T] = self
        return get_enumerator(Queue_1__toSeq(_))

    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None = None) -> IEnumerator[Any]:
        this: Queue_1[_T] = self
        return get_enumerator(this)


Queue_1_reflection = _expr17


def Queue_1__ctor_Z3B4C077E(initial_contents: Array[Any], initial_count: int) -> Queue_1[_T]:
    return Queue_1(initial_contents, initial_count)


def Queue_1__ctor_Z524259A4(initial_capacity: int) -> Queue_1[Any]:
    if initial_capacity < 0:
        raise Exception("capacity is less than 0")

    return Queue_1__ctor_Z3B4C077E(fill([0] * initial_capacity, 0, initial_capacity, None), 0)


def Queue_1__ctor(__unit: None = None) -> Queue_1[Any]:
    return Queue_1__ctor_Z524259A4(4)


def Queue_1__ctor_BB573A(xs: IEnumerable_1[Any]) -> Queue_1[Any]:
    arr: Array[_T] = list(xs)
    return Queue_1__ctor_Z3B4C077E(arr, len(arr))


def Queue_1__get_Count(_: Queue_1[Any]) -> int:
    return _.count


def Queue_1__Enqueue_2B595(_: Queue_1[Any], value: Any) -> None:
    if _.count == Queue_1__size(_):
        Queue_1__ensure_Z524259A4(_, _.count + 1)

    _.contents[_.tail] = value
    _.tail = ((_.tail + 1) % Queue_1__size(_)) or 0
    _.count = (_.count + 1) or 0


def Queue_1__Dequeue(_: Queue_1[Any]) -> Any:
    if _.count == 0:
        raise Exception("Queue is empty")

    value: _T = _.contents[_.head]
    _.head = ((_.head + 1) % Queue_1__size(_)) or 0
    _.count = (_.count - 1) or 0
    return value


def Queue_1__Peek(_: Queue_1[Any]) -> Any:
    if _.count == 0:
        raise Exception("Queue is empty")

    return _.contents[_.head]


def Queue_1__TryDequeue_1F3DB691(this: Queue_1[Any], result: FSharpRef[_T]) -> bool:
    if this.count == 0:
        return False

    else:
        result.contents = Queue_1__Dequeue(this)
        return True


def Queue_1__TryPeek_1F3DB691(this: Queue_1[Any], result: FSharpRef[_T]) -> bool:
    if this.count == 0:
        return False

    else:
        result.contents = Queue_1__Peek(this)
        return True


def Queue_1__Contains_2B595(_: Queue_1[Any], x: Any) -> bool:
    found: bool = False
    i: int = 0
    while (not found) if (i < _.count) else False:
        if equals_1(x, _.contents[Queue_1__toIndex_Z524259A4(_, i)]):
            found = True

        else:
            i = (i + 1) or 0

    return found


def Queue_1__Clear(_: Queue_1[Any]) -> None:
    _.count = 0
    _.head = 0
    _.tail = 0
    fill(_.contents, 0, Queue_1__size(_), None)


def Queue_1__TrimExcess(_: Queue_1[Any]) -> None:
    if divide(_.count, len(_.contents)) > 0.9:
        Queue_1__ensure_Z524259A4(_, _.count)


def Queue_1__ToArray(_: Queue_1[Any]) -> Array[Any]:
    return to_array(Queue_1__toSeq(_))


def Queue_1__CopyTo_Z3B4C077E(_: Queue_1[Any], target: Array[Any], start: int) -> None:
    i: int = start or 0
    with get_enumerator(Queue_1__toSeq(_)) as enumerator:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            item: _T = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            target[i] = item
            i = (i + 1) or 0


def Queue_1__size(this: Queue_1[Any]) -> int:
    return len(this.contents)


def Queue_1__toIndex_Z524259A4(this: Queue_1[Any], i: int) -> int:
    return (this.head + i) % Queue_1__size(this)


def Queue_1__ensure_Z524259A4(this: Queue_1[Any], required_size: int) -> None:
    new_buffer: Array[_T] = fill([0] * required_size, 0, required_size, None)
    if this.head < this.tail:
        copy_to(this.contents, this.head, new_buffer, 0, this.count)

    else:
        copy_to(this.contents, this.head, new_buffer, 0, Queue_1__size(this) - this.head)
        copy_to(this.contents, 0, new_buffer, Queue_1__size(this) - this.head, this.tail)

    this.head = 0
    this.contents = new_buffer
    this.tail = (0 if (this.count == Queue_1__size(this)) else this.count) or 0


def Queue_1__toSeq(this: Queue_1[Any]) -> IEnumerable_1[Any]:
    def _arrow21(__unit: None = None, this: Any = this) -> IEnumerable_1[_T]:
        i: int = 0

        def _arrow18(__unit: None = None) -> bool:
            return i < this.count

        def _arrow20(__unit: None = None) -> IEnumerable_1[_T]:
            def _arrow19(__unit: None = None) -> IEnumerable_1[_T]:
                nonlocal i
                i = (i + 1) or 0
                return empty()

            return append(singleton(this.contents[Queue_1__toIndex_Z524259A4(this, i)]), delay(_arrow19))

        return enumerate_while(_arrow18, delay(_arrow20))

    return delay(_arrow21)


__all__ = [
    "Comparer_1_reflection",
    "Comparer_1_get_Default",
    "Comparer_1_Create_47C913C",
    "Comparer_1__Compare_5BDDA0",
    "EqualityComparer_1_reflection",
    "EqualityComparer_1_get_Default",
    "EqualityComparer_1_Create_Z6EE254AB",
    "EqualityComparer_1__Equals_5BDDA0",
    "EqualityComparer_1__GetHashCode_2B595",
    "Stack_1_reflection",
    "Stack_1__ctor_Z524259A4",
    "Stack_1__ctor",
    "Stack_1__ctor_BB573A",
    "Stack_1__Ensure_Z524259A4",
    "Stack_1__get_Count",
    "Stack_1__Pop",
    "Stack_1__Peek",
    "Stack_1__Contains_2B595",
    "Stack_1__TryPeek_1F3DB691",
    "Stack_1__TryPop_1F3DB691",
    "Stack_1__Push_2B595",
    "Stack_1__Clear",
    "Stack_1__TrimExcess",
    "Stack_1__ToArray",
    "Queue_1_reflection",
    "Queue_1__ctor_Z524259A4",
    "Queue_1__ctor",
    "Queue_1__ctor_BB573A",
    "Queue_1__get_Count",
    "Queue_1__Enqueue_2B595",
    "Queue_1__Dequeue",
    "Queue_1__Peek",
    "Queue_1__TryDequeue_1F3DB691",
    "Queue_1__TryPeek_1F3DB691",
    "Queue_1__Contains_2B595",
    "Queue_1__Clear",
    "Queue_1__TrimExcess",
    "Queue_1__ToArray",
    "Queue_1__CopyTo_Z3B4C077E",
    "Queue_1__size",
    "Queue_1__toIndex_Z524259A4",
    "Queue_1__ensure_Z524259A4",
    "Queue_1__toSeq",
]



from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from ...fable_library.list import (length, empty, FSharpList, choose)
from ...fable_library.option import default_arg
from ...fable_library.reflection import (TypeInfo, string_type, option_type, list_type, record_type)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (to_string, Record)
from ..ontology_annotation import OntologyAnnotation
from .material_attribute_value import (MaterialAttributeValue, MaterialAttributeValue_reflection)

def _expr759() -> TypeInfo:
    return record_type("ARCtrl.Process.Source", [], Source, lambda: [("ID", option_type(string_type)), ("Name", option_type(string_type)), ("Characteristics", option_type(list_type(MaterialAttributeValue_reflection())))])


@dataclass(eq = False, repr = False, slots = True)
class Source(Record):
    ID: str | None
    Name: str | None
    Characteristics: FSharpList[MaterialAttributeValue] | None
    def Print(self, __unit: None=None) -> str:
        this: Source = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: Source = self
        l: int = length(default_arg(this.Characteristics, empty())) or 0
        arg: str = Source__get_NameAsString(this)
        return to_text(printf("%s [%i characteristics]"))(arg)(l)


Source_reflection = _expr759

def Source_make(id: str | None=None, name: str | None=None, characteristics: FSharpList[MaterialAttributeValue] | None=None) -> Source:
    return Source(id, name, characteristics)


def Source_create_Z5CA08497(Id: str | None=None, Name: str | None=None, Characteristics: FSharpList[MaterialAttributeValue] | None=None) -> Source:
    return Source_make(Id, Name, Characteristics)


def Source_get_empty(__unit: None=None) -> Source:
    return Source_create_Z5CA08497()


def Source__get_NameAsString(this: Source) -> str:
    return default_arg(this.Name, "")


def Source_getUnits_53E41069(m: Source) -> FSharpList[OntologyAnnotation]:
    def chooser(c: MaterialAttributeValue, m: Any=m) -> OntologyAnnotation | None:
        return c.Unit

    return choose(chooser, default_arg(m.Characteristics, empty()))


def Source_setCharacteristicValues(values: FSharpList[MaterialAttributeValue], m: Source) -> Source:
    return Source(m.ID, m.Name, values)


__all__ = ["Source_reflection", "Source_make", "Source_create_Z5CA08497", "Source_get_empty", "Source__get_NameAsString", "Source_getUnits_53E41069", "Source_setCharacteristicValues"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.map import of_seq
from ..fable_library.option import value
from ..fable_library.seq import (try_pick, exists)
from ..fable_library.types import Array
from ..fable_library.util import (compare_primitives, equals)
from .comment import Comment
from .Helper.collections_ import (ResizeArray_choose, ResizeArray_map, ResizeArray_appendSingleton, ResizeArray_filter)

def try_item(key: str, comments: Array[Comment]) -> str | None:
    def chooser(c: Comment, key: Any=key, comments: Any=comments) -> str | None:
        match_value: str | None = c.Name
        (pattern_matching_result,) = (None,)
        if match_value is not None:
            if match_value == key:
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return c.Value

        elif pattern_matching_result == 1:
            return None


    return try_pick(chooser, comments)


def contains_key(key: str, comments: Array[Comment]) -> bool:
    def predicate(c: Comment, key: Any=key, comments: Any=comments) -> bool:
        match_value: str | None = c.Name
        (pattern_matching_result,) = (None,)
        if match_value is not None:
            if match_value == key:
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return True

        elif pattern_matching_result == 1:
            return False


    return exists(predicate, comments)


def item(key: str, comments: Array[Comment]) -> str:
    return value(try_item(key, comments))


def to_map(comments: Array[Comment]) -> Any:
    def f(c: Comment, comments: Any=comments) -> tuple[str, str | None] | None:
        match_value: str | None = c.Name
        if match_value is not None:
            return (match_value, c.Value)

        else: 
            return None


    class ObjectExpr699:
        @property
        def Compare(self) -> Callable[[str, str], int]:
            return compare_primitives

    return of_seq(ResizeArray_choose(f, comments), ObjectExpr699())


def set_1(comment: Comment, comments: Array[Comment]) -> Array[Comment]:
    if contains_key(value(comment.Name), comments):
        def f(c: Comment, comment: Any=comment, comments: Any=comments) -> Comment:
            if equals(c.Name, comment.Name):
                return comment

            else: 
                return c


        return ResizeArray_map(f, comments)

    else: 
        return ResizeArray_appendSingleton(comment, comments)



def drop_by_key(key: str, comments: Array[Comment]) -> Array[Comment]:
    def f(c: Comment, key: Any=key, comments: Any=comments) -> bool:
        match_value: str | None = c.Name
        (pattern_matching_result,) = (None,)
        if match_value is not None:
            if match_value == key:
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return False

        elif pattern_matching_result == 1:
            return True


    return ResizeArray_filter(f, comments)


__all__ = ["try_item", "contains_key", "item", "to_map", "set_1", "drop_by_key"]




import inspect
from abc import abstractmethod
from collections.abc import Callable
from typing import (
    Any,
    Generic,
    Protocol,
    TypeVar,
    overload,
)

from .choice import (
    Choice_tryValueIfChoice1Of2,
    Choice_tryValueIfChoice2Of2,
    FSharpChoice_2,
)
from .observable import IObservable, IObserver, Observer
from .option import some, value
from .util import IDisposable


_T = TypeVar("_T")
_U = TypeVar("_U")
_V = TypeVar("_V")
_Args = TypeVar("_Args", covariant=True)
_Delegate = TypeVar("_Delegate", covariant=True)
_T_co = TypeVar("_T_co", covariant=True)

Delegate = Callable[[_T], None]
DotNetDelegate = Callable[[Any, _T], None]
EventDelegate = Delegate[_T] | DotNetDelegate[_T]


class IDelegateEvent(Generic[_T_co], Protocol):
    @abstractmethod
    def AddHandler(self, d: DotNetDelegate[_T]) -> None: ...

    @abstractmethod
    def RemoveHandler(self, d: DotNetDelegate[_T]) -> None: ...


class IEvent_2(IObservable[_Args], IDelegateEvent[_Delegate], Protocol): ...


IEvent = IEvent_2[_T, _T]


class Event(IEvent[_T]):
    def __init__(self) -> None:
        self.delegates: list[EventDelegate[_T]] = []

    def Add(self, f: Delegate[_T]) -> None:
        self._addHandler(f)

    @property
    def Publish(self) -> IEvent[_T]:
        return self

    @overload
    def Trigger(self, value: _T) -> None: ...

    @overload
    def Trigger(self, sender: Any, value: _T) -> None: ...

    def Trigger(self, sender_or_value: Any, value_or_undefined: _T | None = None) -> None:
        if value_or_undefined is None:
            value = sender_or_value
            sender = None
        else:
            sender = sender_or_value
            value = value_or_undefined

        for f in self.delegates:
            if len(inspect.signature(f).parameters) == 1:
                f(value)
            else:
                f(sender, value)

    # IDelegateEvent<T> methods

    def AddHandler(self, handler: DotNetDelegate[_T]) -> None:
        self._addHandler(handler)

    def RemoveHandler(self, handler: DotNetDelegate[_T]) -> None:
        self._removeHandler(handler)

    # IObservable<T> methods

    def Subscribe(self, __obs: IObserver[_T] | Delegate[_T]) -> IDisposable:
        if callable(__obs):
            callback = __obs
        else:
            callback = __obs.OnNext

        self._addHandler(callback)

        def dispose() -> None:
            self._removeHandler(callback)

        return IDisposable.create(dispose)

    def _addHandler(self, f: Delegate[_T]) -> None:
        self.delegates.append(f)

    def _removeHandler(self, f: EventDelegate[_T]) -> None:
        if f in self.delegates:
            self.delegates.remove(f)


def add(callback: Delegate[_T], source_event: IEvent[_T]) -> None:
    if isinstance(source_event, Event):
        source_event.Add(callback)
    else:
        source_event.Subscribe(Observer(callback))


def choose(chooser: Callable[[_T], _U | None], source_event: IEvent[_T]) -> IEvent[_U]:
    ev = Event[_U]()

    def callback(t):
        u = chooser(t)
        if u is not None:
            ev.Trigger(value(u))

    add(callback, source_event)
    return ev


def filter(predicate: Callable[[_T], bool], source_event: IEvent[_T]) -> IEvent[_T]:
    return choose(lambda t: some(t) if predicate(t) else None, source_event)


def map(mapping: Callable[[_T], _U], source_event: IEvent[_T]) -> IEvent[_U]:
    return choose(lambda t: mapping(t), source_event)


def merge(event1: IEvent[_T], event2: IEvent[_T]) -> IEvent[_T]:
    ev = Event[_T]()

    def fn(t):
        ev.Trigger(t)

    add(fn, event1)
    add(fn, event2)
    return ev


def pairwise(source_event: IEvent[_T]) -> IEvent[list[_T]]:
    ev = Event[list[_T]]()
    last = None

    def fn(next):
        nonlocal last
        if last is not None:
            ev.Trigger([last, next])
        last = next

    add(fn, source_event)
    return ev


def partition(predicate: Callable[[_T], bool], source_event: IEvent[_T]) -> list[IEvent[_T]]:
    return [
        filter(predicate, source_event),
        filter(lambda x: not predicate(x), source_event),
    ]


def scan(
    collector: Callable[[_U, _T], _U],
    state: _U,
    source_event: IEvent[_T],
) -> IEvent[_U]:
    return map(lambda t: collector(state, t), source_event)


def split(splitter: Callable[[_T], FSharpChoice_2[_U, _V]], source_event: IEvent[_T]) -> list[IEvent[_U | _V]]:
    return [
        choose(lambda t: Choice_tryValueIfChoice1Of2(splitter(t)), source_event),
        choose(lambda t: Choice_tryValueIfChoice2Of2(splitter(t)), source_event),
    ]


def create_event(
    add_handler: Callable[[DotNetDelegate[_T]], None],
    remove_handler: Callable[[DotNetDelegate[_T]], None],
) -> IEvent[_T]:
    class AnonymousEvent(IEvent[_U]):
        def AddHandler(self, handler: DotNetDelegate[_T]) -> None:
            add_handler(handler)

        def RemoveHandler(self, handler: DotNetDelegate[_T]) -> None:
            remove_handler(handler)

        def Subscribe(self, observer: IObserver[_T]) -> IDisposable:
            def h(sender, value):
                observer.OnNext(value)

            add_handler(h)

            def dispose() -> None:
                remove_handler(h)

            return IDisposable.create(dispose)

    return AnonymousEvent[_T]()



from __future__ import annotations
from typing import Any
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from .json import Json
from .xlsx import Xlsx

def FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxFile_Static_Z721C83C5(path: str) -> FsWorkbook:
    return Xlsx.from_xlsx_file(path)


def FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxStream_Static_4D976C1A(stream: Any) -> FsWorkbook:
    return Xlsx.from_xlsx_stream(stream)


def FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxBytes_Static_Z3F6BC7B1(bytes: bytearray) -> FsWorkbook:
    return Xlsx.from_xlsx_bytes(bytes)


def FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static(path: str, wb: FsWorkbook) -> None:
    Xlsx.to_xlsx_file(path, wb)


def FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxBytes_Static_32154C9D(wb: FsWorkbook) -> bytearray:
    return Xlsx.to_xlsx_bytes(wb)


def FsSpreadsheet_FsWorkbook__FsWorkbook_ToXlsxFile_Z721C83C5(this: FsWorkbook, path: str) -> None:
    FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static(path, this)


def FsSpreadsheet_FsWorkbook__FsWorkbook_ToXlsxBytes(this: FsWorkbook) -> bytearray:
    return FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxBytes_Static_32154C9D(this)


def FsSpreadsheet_FsWorkbook__FsWorkbook_fromRowsJsonString_Static_Z721C83C5(json: str) -> FsWorkbook:
    return Json.from_rows_json_string(json)


def FsSpreadsheet_FsWorkbook__FsWorkbook_toRowsJsonString_Static_Z2B6E0EF5(wb: FsWorkbook, spaces: int | None=None, no_numbering: bool | None=None) -> str:
    return Json.to_rows_json_string(wb, spaces, no_numbering)


def FsSpreadsheet_FsWorkbook__FsWorkbook_ToRowsJsonString_Z3B036AA(this: FsWorkbook, spaces: int | None=None, no_numbering: bool | None=None) -> str:
    return FsSpreadsheet_FsWorkbook__FsWorkbook_toRowsJsonString_Static_Z2B6E0EF5(this, spaces, no_numbering)


def FsSpreadsheet_FsWorkbook__FsWorkbook_fromColumnsJsonString_Static_Z721C83C5(json: str) -> FsWorkbook:
    return Json.from_columns_json_string(json)


def FsSpreadsheet_FsWorkbook__FsWorkbook_toColumnsJsonString_Static_Z2B6E0EF5(wb: FsWorkbook, spaces: int | None=None, no_numbering: bool | None=None) -> str:
    return Json.to_columns_json_string(wb, spaces, no_numbering)


def FsSpreadsheet_FsWorkbook__FsWorkbook_ToColumnsJsonString_Z3B036AA(this: FsWorkbook, spaces: int | None=None, no_numbering: bool | None=None) -> str:
    return FsSpreadsheet_FsWorkbook__FsWorkbook_toColumnsJsonString_Static_Z2B6E0EF5(this, spaces, no_numbering)


__all__ = ["FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxFile_Static_Z721C83C5", "FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxStream_Static_4D976C1A", "FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxBytes_Static_Z3F6BC7B1", "FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static", "FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxBytes_Static_32154C9D", "FsSpreadsheet_FsWorkbook__FsWorkbook_ToXlsxFile_Z721C83C5", "FsSpreadsheet_FsWorkbook__FsWorkbook_ToXlsxBytes", "FsSpreadsheet_FsWorkbook__FsWorkbook_fromRowsJsonString_Static_Z721C83C5", "FsSpreadsheet_FsWorkbook__FsWorkbook_toRowsJsonString_Static_Z2B6E0EF5", "FsSpreadsheet_FsWorkbook__FsWorkbook_ToRowsJsonString_Z3B036AA", "FsSpreadsheet_FsWorkbook__FsWorkbook_fromColumnsJsonString_Static_Z721C83C5", "FsSpreadsheet_FsWorkbook__FsWorkbook_toColumnsJsonString_Static_Z2B6E0EF5", "FsSpreadsheet_FsWorkbook__FsWorkbook_ToColumnsJsonString_Z3B036AA"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ..arctrl_core.comment import Comment
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.publication import Publication
from ..fable_library.list import (choose, of_array, FSharpList)
from ..fable_library.option import map
from ..fable_library.seq import map as map_1
from ..fable_library.string_ import replace
from ..fable_library.types import Array
from ..fable_library.util import IEnumerable_1
from ..thoth_json_core.decode import (object, IOptionalGetter, string, resize_array, IGetters)
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .comment import (encoder as encoder_1, decoder as decoder_2, ROCrate_encoderDisambiguatingDescription, ROCrate_decoderDisambiguatingDescription, ISAJson_encoder as ISAJson_encoder_1)
from .context.rocrate.isa_publication_context import context_jsonvalue
from .decode import (Decode_uri, Decode_noAdditionalProperties)
from .encode import (try_include, try_include_seq)
from .ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder, OntologyAnnotation_ROCrate_encoderDefinedTerm, OntologyAnnotation_ROCrate_decoderDefinedTerm)
from .person import (ROCrate_encodeAuthorListString, ROCrate_decodeAuthorListString)

__A_ = TypeVar("__A_")

def encoder(oa: Publication) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2240(value: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2239(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2239()

    def _arrow2242(value_2: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2241(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_2)

        return ObjectExpr2241()

    def _arrow2244(value_4: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2243(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_4)

        return ObjectExpr2243()

    def _arrow2246(value_6: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2245(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_6)

        return ObjectExpr2245()

    def _arrow2247(oa_1: OntologyAnnotation, oa: Any=oa) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow2248(comment: Comment, oa: Any=oa) -> IEncodable:
        return encoder_1(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("pubMedID", _arrow2240, oa.PubMedID), try_include("doi", _arrow2242, oa.DOI), try_include("authorList", _arrow2244, oa.Authors), try_include("title", _arrow2246, oa.Title), try_include("status", _arrow2247, oa.Status), try_include_seq("comments", _arrow2248, oa.Comments)]))
    class ObjectExpr2249(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr2249()


def _arrow2256(get: IGetters) -> Publication:
    def _arrow2250(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("pubMedID", Decode_uri)

    def _arrow2251(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("doi", string)

    def _arrow2252(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("authorList", string)

    def _arrow2253(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("title", string)

    def _arrow2254(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("status", OntologyAnnotation_decoder)

    def _arrow2255(__unit: None=None) -> Array[Comment] | None:
        arg_11: Decoder_1[Array[Comment]] = resize_array(decoder_2)
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("comments", arg_11)

    return Publication(_arrow2250(), _arrow2251(), _arrow2252(), _arrow2253(), _arrow2254(), _arrow2255())


decoder: Decoder_1[Publication] = object(_arrow2256)

def ROCrate_genID(p: Publication) -> str:
    match_value: str | None = p.DOI
    if match_value is None:
        match_value_1: str | None = p.PubMedID
        if match_value_1 is None:
            match_value_2: str | None = p.Title
            if match_value_2 is None:
                return "#EmptyPublication"

            else: 
                return "#Pub_" + replace(match_value_2, " ", "_")


        else: 
            return match_value_1


    else: 
        return match_value



def ROCrate_encoder(oa: Publication) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2260(__unit: None=None, oa: Any=oa) -> IEncodable:
        value: str = ROCrate_genID(oa)
        class ObjectExpr2259(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2259()

    class ObjectExpr2261(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            return helpers_1.encode_string("Publication")

    def _arrow2263(value_2: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2262(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_2)

        return ObjectExpr2262()

    def _arrow2265(value_4: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2264(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_4)

        return ObjectExpr2264()

    def _arrow2266(author_list: str, oa: Any=oa) -> IEncodable:
        return ROCrate_encodeAuthorListString(author_list)

    def _arrow2268(value_6: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2267(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_6)

        return ObjectExpr2267()

    def _arrow2269(oa_1: OntologyAnnotation, oa: Any=oa) -> IEncodable:
        return OntologyAnnotation_ROCrate_encoderDefinedTerm(oa_1)

    def _arrow2270(comment: Comment, oa: Any=oa) -> IEncodable:
        return ROCrate_encoderDisambiguatingDescription(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2260()), ("@type", ObjectExpr2261()), try_include("pubMedID", _arrow2263, oa.PubMedID), try_include("doi", _arrow2265, oa.DOI), try_include("authorList", _arrow2266, oa.Authors), try_include("title", _arrow2268, oa.Title), try_include("status", _arrow2269, oa.Status), try_include_seq("comments", _arrow2270, oa.Comments), ("@context", context_jsonvalue)]))
    class ObjectExpr2271(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_5))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_5.encode_object(arg)

    return ObjectExpr2271()


def _arrow2278(get: IGetters) -> Publication:
    def _arrow2272(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("pubMedID", Decode_uri)

    def _arrow2273(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("doi", string)

    def _arrow2274(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("authorList", ROCrate_decodeAuthorListString)

    def _arrow2275(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("title", string)

    def _arrow2276(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("status", OntologyAnnotation_ROCrate_decoderDefinedTerm)

    def _arrow2277(__unit: None=None) -> Array[Comment] | None:
        arg_11: Decoder_1[Array[Comment]] = resize_array(ROCrate_decoderDisambiguatingDescription)
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("comments", arg_11)

    return Publication(_arrow2272(), _arrow2273(), _arrow2274(), _arrow2275(), _arrow2276(), _arrow2277())


ROCrate_decoder: Decoder_1[Publication] = object(_arrow2278)

def ISAJson_encoder(id_map: Any | None, oa: Publication) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], id_map: Any=id_map, oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2282(value: str, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        class ObjectExpr2281(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2281()

    def _arrow2284(value_2: str, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        class ObjectExpr2283(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_2)

        return ObjectExpr2283()

    def _arrow2286(value_4: str, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        class ObjectExpr2285(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_4)

        return ObjectExpr2285()

    def _arrow2290(value_6: str, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        class ObjectExpr2289(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_6)

        return ObjectExpr2289()

    def _arrow2291(oa_1: OntologyAnnotation, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow2292(comment: Comment, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        return ISAJson_encoder_1(id_map, comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("pubMedID", _arrow2282, oa.PubMedID), try_include("doi", _arrow2284, oa.DOI), try_include("authorList", _arrow2286, oa.Authors), try_include("title", _arrow2290, oa.Title), try_include("status", _arrow2291, oa.Status), try_include_seq("comments", _arrow2292, oa.Comments)]))
    class ObjectExpr2297(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], id_map: Any=id_map, oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr2297()


ISAJson_allowedFields: FSharpList[str] = of_array(["pubMedID", "doi", "authorList", "title", "status", "comments"])

ISAJson_decoder: Decoder_1[Publication] = Decode_noAdditionalProperties(ISAJson_allowedFields, decoder)

__all__ = ["encoder", "decoder", "ROCrate_genID", "ROCrate_encoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_allowedFields", "ISAJson_decoder"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.Helper.identifier import create_missing_identifier
from ...fable_library.list import (append as append_1, singleton, FSharpList, is_empty)
from ...fable_library.option import value
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import (append, to_list, delay, singleton as singleton_1, empty)
from ...fable_library.string_ import join
from ...fable_library.types import Array
from ...fable_library.util import IEnumerable_1
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph)
from .comment import LDComment
from .defined_term import LDDefinedTerm

def _expr1524() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDLabProtocol", None, LDLabProtocol)


class LDLabProtocol:
    @staticmethod
    def schema_type() -> str:
        return "https://bioschemas.org/LabProtocol"

    @staticmethod
    def description() -> str:
        return "http://schema.org/description"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def computational_tool() -> str:
        return "https://bioschemas.org/properties/computationalTool"

    @staticmethod
    def computational_tool_deprecated() -> str:
        return "https://bioschemas.org/computationalTool"

    @staticmethod
    def lab_equipment() -> str:
        return "https://bioschemas.org/properties/labEquipment"

    @staticmethod
    def lab_equipment_deprecated() -> str:
        return "https://bioschemas.org/labEquipment"

    @staticmethod
    def reagent() -> str:
        return "https://bioschemas.org/properties/reagent"

    @staticmethod
    def reagent_deprecated() -> str:
        return "https://bioschemas.org/reagent"

    @staticmethod
    def intended_use() -> str:
        return "https://bioschemas.org/properties/intendedUse"

    @staticmethod
    def intended_use_deprecated() -> str:
        return "https://bioschemas.org/intendedUse"

    @staticmethod
    def url() -> str:
        return "http://schema.org/url"

    @staticmethod
    def version() -> str:
        return "http://schema.org/version"

    @staticmethod
    def comment() -> str:
        return "http://schema.org/comment"

    @staticmethod
    def try_get_description_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDLabProtocol.description(), context)
        (pattern_matching_result, d) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                d = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return d

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_description_as_string(lp: LDNode, description: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProtocol.description(), description, context)

    @staticmethod
    def try_get_intended_use_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDLabProtocol.intended_use(), context)
        (pattern_matching_result, iu) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                iu = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return iu

        elif pattern_matching_result == 1:
            match_value_1: Any | None = lp.TryGetPropertyAsSingleton(LDLabProtocol.intended_use_deprecated(), context)
            (pattern_matching_result_1, iu_1) = (None, None)
            if match_value_1 is not None:
                if str(type(value(match_value_1))) == "<class \'str\'>":
                    pattern_matching_result_1 = 0
                    iu_1 = value(match_value_1)

                else: 
                    pattern_matching_result_1 = 1


            else: 
                pattern_matching_result_1 = 1

            if pattern_matching_result_1 == 0:
                return iu_1

            elif pattern_matching_result_1 == 1:
                return None



    @staticmethod
    def set_intended_use_as_string(lp: LDNode, intended_use: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProtocol.intended_use(), intended_use, context)

    @staticmethod
    def try_get_intended_use_as_defined_term(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDDefinedTerm.validate(ld_object, context_1)

        match_value: LDNode | None = lp.TryGetPropertyAsSingleNode(LDLabProtocol.intended_use(), graph, context)
        (pattern_matching_result, iu_1) = (None, None)
        if match_value is not None:
            if filter(match_value, context):
                pattern_matching_result = 0
                iu_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return iu_1

        elif pattern_matching_result == 1:
            match_value_1: LDNode | None = lp.TryGetPropertyAsSingleNode(LDLabProtocol.intended_use_deprecated(), graph, context)
            (pattern_matching_result_1, iu_3) = (None, None)
            if match_value_1 is not None:
                if filter(match_value_1, context):
                    pattern_matching_result_1 = 0
                    iu_3 = match_value_1

                else: 
                    pattern_matching_result_1 = 1


            else: 
                pattern_matching_result_1 = 1

            if pattern_matching_result_1 == 0:
                return iu_3

            elif pattern_matching_result_1 == 1:
                return None



    @staticmethod
    def set_intended_use_as_defined_term(lp: LDNode, intended_use: LDNode, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProtocol.intended_use(), intended_use, context)

    @staticmethod
    def try_get_name_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDLabProtocol.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(lp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDLabProtocol.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            raise Exception(("Could not access property `name` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_name_as_string(lp: LDNode, name: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProtocol.name(), name, context)

    @staticmethod
    def get_comments(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDComment.validate(ld_object, context_1)

        return lp.GetPropertyNodes(LDLabProtocol.comment(), filter, graph, context)

    @staticmethod
    def set_comments(lp: LDNode, comments: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProtocol.comment(), comments, context)

    @staticmethod
    def get_computational_tools(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        l: Array[LDNode] = lp.GetPropertyNodes(LDLabProtocol.computational_tool(), None, graph, context)
        return lp.GetPropertyNodes(LDLabProtocol.computational_tool_deprecated(), None, graph, context) if (len(l) == 0) else l

    @staticmethod
    def set_computational_tools(lp: LDNode, computational_tools: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProtocol.computational_tool(), computational_tools, context)

    @staticmethod
    def get_lab_equipments(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        l: Array[LDNode] = lp.GetPropertyNodes(LDLabProtocol.lab_equipment(), None, graph, context)
        return lp.GetPropertyNodes(LDLabProtocol.lab_equipment_deprecated(), None, graph, context) if (len(l) == 0) else l

    @staticmethod
    def set_lab_equipments(lp: LDNode, lab_equipments: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProtocol.lab_equipment(), lab_equipments, context)

    @staticmethod
    def get_reagents(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        l: Array[LDNode] = lp.GetPropertyNodes(LDLabProtocol.reagent(), None, graph, context)
        return lp.GetPropertyNodes(LDLabProtocol.reagent_deprecated(), None, graph, context) if (len(l) == 0) else l

    @staticmethod
    def set_reagents(lp: LDNode, reagents: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProtocol.reagent(), reagents, context)

    @staticmethod
    def get_components(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def _arrow1518(__unit: None=None) -> IEnumerable_1[LDNode]:
            source_3: IEnumerable_1[LDNode]
            source_1: Array[LDNode] = LDLabProtocol.get_lab_equipments(lp, graph, context)
            source_3 = append(LDLabProtocol.get_reagents(lp, graph, context), source_1)
            return append(LDLabProtocol.get_computational_tools(lp, graph, context), source_3)

        return list(_arrow1518())

    @staticmethod
    def try_get_url(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDLabProtocol.url(), context)
        (pattern_matching_result, u) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                u = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return u

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_url(lp: LDNode, url: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProtocol.url(), url, context)

    @staticmethod
    def try_get_version_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDLabProtocol.version(), context)
        (pattern_matching_result, v) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                v = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return v

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_version_as_string(lp: LDNode, version: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDLabProtocol.version(), version, context)

    @staticmethod
    def validate(lp: LDNode, context: LDContext | None=None) -> bool:
        return lp.HasType(LDLabProtocol.schema_type(), context)

    @staticmethod
    def gen_id(name: str | None=None, process_name: str | None=None, assay_name: str | None=None, study_name: str | None=None) -> str:
        def _arrow1523(__unit: None=None) -> FSharpList[str]:
            def _arrow1522(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow1521(__unit: None=None) -> IEnumerable_1[str]:
                    def _arrow1520(__unit: None=None) -> IEnumerable_1[str]:
                        def _arrow1519(__unit: None=None) -> IEnumerable_1[str]:
                            return singleton_1(value(study_name)) if (study_name is not None) else empty()

                        return append(singleton_1(value(assay_name)) if (assay_name is not None) else empty(), delay(_arrow1519))

                    return append(singleton_1(value(process_name)) if (process_name is not None) else empty(), delay(_arrow1520))

                return append(singleton_1(value(name)) if (name is not None) else empty(), delay(_arrow1521))

            vals: FSharpList[str] = to_list(delay(_arrow1522))
            return singleton(create_missing_identifier()) if is_empty(vals) else vals

        return clean(join("_", append_1(singleton("#Protocol"), _arrow1523())))

    @staticmethod
    def create(id: str, name: str | None=None, description: str | None=None, intended_use: LDNode | None=None, comments: Array[LDNode] | None=None, computational_tools: Array[LDNode] | None=None, lab_equipments: Array[LDNode] | None=None, reagents: Array[LDNode] | None=None, url: str | None=None, version: str | None=None, context: LDContext | None=None) -> LDNode:
        lp: LDNode = LDNode(id, [LDLabProtocol.schema_type()], None, context)
        lp.SetOptionalProperty(LDLabProtocol.name(), name, context)
        lp.SetOptionalProperty(LDLabProtocol.description(), description, context)
        lp.SetOptionalProperty(LDLabProtocol.intended_use(), intended_use, context)
        lp.SetOptionalProperty(LDLabProtocol.comment(), comments, context)
        lp.SetOptionalProperty(LDLabProtocol.computational_tool(), computational_tools, context)
        lp.SetOptionalProperty(LDLabProtocol.lab_equipment(), lab_equipments, context)
        lp.SetOptionalProperty(LDLabProtocol.reagent(), reagents, context)
        lp.SetOptionalProperty(LDLabProtocol.url(), url, context)
        lp.SetOptionalProperty(LDLabProtocol.version(), version, context)
        return lp


LDLabProtocol_reflection = _expr1524

__all__ = ["LDLabProtocol_reflection"]




from __future__ import annotations
from typing import Any
from ...fable_library.list import (FSharpList, singleton)
from ...fable_library.reflection import (TypeInfo, class_type)
from ..Cells.fs_cell import DataType
from .types import (SheetEntity_1, Message, RowElement, ColumnElement, SheetElement, WorkbookElement)

def _expr322() -> TypeInfo:
    return class_type("FsSpreadsheet.DSL.DSL", None, DSL)


class DSL:
    ...

DSL_reflection = _expr322

def DSL_opt_Z6AB9374C(elem: SheetEntity_1[FSharpList[Any]]) -> SheetEntity_1[FSharpList[Any]]:
    if elem.tag == 1:
        return SheetEntity_1(1, elem.fields[0])

    elif elem.tag == 2:
        return SheetEntity_1(1, elem.fields[0])

    else: 
        return elem



def DSL_dropCell_6CC5727E(message: Message) -> SheetEntity_1[tuple[DataType, Any]]:
    return SheetEntity_1(2, singleton(message))


def DSL_dropRow_6CC5727E(message: Message) -> SheetEntity_1[RowElement]:
    return SheetEntity_1(2, singleton(message))


def DSL_dropColumn_6CC5727E(message: Message) -> SheetEntity_1[ColumnElement]:
    return SheetEntity_1(2, singleton(message))


def DSL_dropSheet_6CC5727E(message: Message) -> SheetEntity_1[SheetElement]:
    return SheetEntity_1(2, singleton(message))


def DSL_dropWorkbook_6CC5727E(message: Message) -> SheetEntity_1[WorkbookElement]:
    return SheetEntity_1(2, singleton(message))


__all__ = ["DSL_reflection", "DSL_opt_Z6AB9374C", "DSL_dropCell_6CC5727E", "DSL_dropRow_6CC5727E", "DSL_dropColumn_6CC5727E", "DSL_dropSheet_6CC5727E", "DSL_dropWorkbook_6CC5727E"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.Helper.collections_ import ResizeArray_map
from ...fable_library.array_ import contains
from ...fable_library.int32 import try_parse
from ...fable_library.option import (value as value_3, default_arg, to_array)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import iterate
from ...fable_library.string_ import (join, replace)
from ...fable_library.types import (Array, FSharpRef)
from ...fable_library.util import (string_hash, int32_to_string)
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph)
from .formal_parameter import LDFormalParameter

def _expr1500() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDPropertyValue", None, LDPropertyValue)


class LDPropertyValue:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/PropertyValue"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def value() -> str:
        return "http://schema.org/value"

    @staticmethod
    def property_id() -> str:
        return "http://schema.org/propertyID"

    @staticmethod
    def unit_code() -> str:
        return "http://schema.org/unitCode"

    @staticmethod
    def unit_text() -> str:
        return "http://schema.org/unitText"

    @staticmethod
    def value_reference() -> str:
        return "http://schema.org/valueReference"

    @staticmethod
    def measurement_method() -> str:
        return "http://schema.org/measurementMethod"

    @staticmethod
    def description() -> str:
        return "http://schema.org/description"

    @staticmethod
    def alternate_name() -> str:
        return "http://schema.org/alternateName"

    @staticmethod
    def subject_of() -> str:
        return "http://schema.org/subjectOf"

    @staticmethod
    def disambiguating_description() -> str:
        return "http://schema.org/disambiguatingDescription"

    @staticmethod
    def doi_key() -> str:
        return "DOI"

    @staticmethod
    def doi_url() -> str:
        return "http://purl.obolibrary.org/obo/OBI_0002110"

    @staticmethod
    def pubmed_idkey() -> str:
        return "PubMedID"

    @staticmethod
    def pubmed_idurl() -> str:
        return "http://purl.obolibrary.org/obo/OBI_0001617"

    @staticmethod
    def prefix_key() -> str:
        return "Prefix"

    @staticmethod
    def position_key() -> str:
        return "Position"

    @staticmethod
    def glob_key() -> str:
        return "Glob"

    @staticmethod
    def example_of_work() -> str:
        return "http://schema.org/exampleOfWork"

    @staticmethod
    def try_get_name_as_string(pv: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = pv.TryGetPropertyAsSingleton(LDPropertyValue.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_3(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(pv: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = pv.TryGetPropertyAsSingleton(LDPropertyValue.name(), context)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                n: str = value_3(match_value)
                return n

            else: 
                raise Exception(("Property of `name` of object with @id `" + pv.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `name` of object with @id `" + pv.Id) + "`")


    @staticmethod
    def set_name_as_string(pv: LDNode, name: str, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDPropertyValue.name(), name, context)

    @staticmethod
    def get_values_as_string(pv: LDNode, context: LDContext | None=None) -> Array[str]:
        def f(o_1: Any=None) -> Any:
            return o_1

        def filter(o: Any=None, context_1: LDContext | None=None) -> bool:
            return str(type(o)) == "<class \'str\'>"

        return ResizeArray_map(f, pv.GetPropertyValues(LDPropertyValue.value(), filter, context))

    @staticmethod
    def try_get_value_as_string(pv: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = pv.TryGetPropertyAsSingleton(LDPropertyValue.value(), context)
        (pattern_matching_result, v) = (None, None)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                v = value_3(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return v

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_value_as_string(pv: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = pv.TryGetPropertyAsSingleton(LDPropertyValue.value(), context)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                v: str = value_3(match_value)
                return v

            else: 
                raise Exception(("Property of `value` of object with @id `" + pv.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `value` of object with @id `" + pv.Id) + "`")


    @staticmethod
    def set_value_as_string(pv: LDNode, value: str, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDPropertyValue.value(), value, context)

    @staticmethod
    def try_get_property_idas_string(pv: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = pv.TryGetPropertyAsSingleton(LDPropertyValue.property_id(), context)
        (pattern_matching_result, pid) = (None, None)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                pid = value_3(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return pid

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_property_idas_string(pv: LDNode, property_id: str, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDPropertyValue.property_id(), property_id, context)

    @staticmethod
    def try_get_unit_code_as_string(pv: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = pv.TryGetPropertyAsSingleton(LDPropertyValue.unit_code(), context)
        (pattern_matching_result, uc) = (None, None)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                uc = value_3(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return uc

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_unit_code_as_string(pv: LDNode, unit_code: str, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDPropertyValue.unit_code(), unit_code, context)

    @staticmethod
    def try_get_unit_text_as_string(pv: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = pv.TryGetPropertyAsSingleton(LDPropertyValue.unit_text(), context)
        (pattern_matching_result, ut) = (None, None)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                ut = value_3(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return ut

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_unit_text_as_string(pv: LDNode, unit_text: str, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDPropertyValue.unit_text(), unit_text, context)

    @staticmethod
    def try_get_value_reference_as_string(pv: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = pv.TryGetPropertyAsSingleton(LDPropertyValue.value_reference(), context)
        (pattern_matching_result, vr) = (None, None)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                vr = value_3(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return vr

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_value_reference_as_string(pv: LDNode, value_reference: str, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDPropertyValue.value_reference(), value_reference, context)

    @staticmethod
    def try_get_measurement_method_as_string(pv: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = pv.TryGetPropertyAsSingleton(LDPropertyValue.measurement_method(), context)
        (pattern_matching_result, vr) = (None, None)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                vr = value_3(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return vr

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_measurement_method_as_string(pv: LDNode, measurement_method: str, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDPropertyValue.measurement_method(), measurement_method, context)

    @staticmethod
    def try_get_description_as_string(pv: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = pv.TryGetPropertyAsSingleton(LDPropertyValue.description(), context)
        (pattern_matching_result, vr) = (None, None)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                vr = value_3(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return vr

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_description_as_string(pv: LDNode, description: str, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDPropertyValue.description(), description, context)

    @staticmethod
    def try_get_alternate_name_as_string(pv: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = pv.TryGetPropertyAsSingleton(LDPropertyValue.alternate_name(), context)
        (pattern_matching_result, vr) = (None, None)
        if match_value is not None:
            if str(type(value_3(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                vr = value_3(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return vr

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_alternate_name_as_string(pv: LDNode, alternate_name: str, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDPropertyValue.alternate_name(), alternate_name, context)

    @staticmethod
    def get_disambiguating_descriptions_as_string(pv: LDNode, context: LDContext | None=None) -> Array[str]:
        def f(o_1: Any=None) -> Any:
            return o_1

        def filter(o: Any=None, context_1: LDContext | None=None) -> bool:
            return str(type(o)) == "<class \'str\'>"

        return ResizeArray_map(f, pv.GetPropertyValues(LDPropertyValue.disambiguating_description(), filter, context))

    @staticmethod
    def set_disambiguating_descriptions_as_string(lp: LDNode, disambiguating_descriptions: Array[str], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDPropertyValue.disambiguating_description(), disambiguating_descriptions, context)

    @staticmethod
    def try_get_subject_of(pv: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = pv.TryGetPropertyAsSingleNode(LDPropertyValue.subject_of(), graph, context)
        return match_value if (match_value is not None) else None

    @staticmethod
    def set_subject_of(pv: LDNode, subject_of: LDNode, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDPropertyValue.subject_of(), subject_of, context)

    @staticmethod
    def try_get_example_of_work(dt: LDNode, context: LDContext | None=None) -> Any | None:
        return dt.TryGetPropertyAsSingleton(LDPropertyValue.example_of_work(), context)

    @staticmethod
    def try_get_example_of_work_as_formal_parameter(dt: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = dt.TryGetPropertyAsSingleNode(LDPropertyValue.example_of_work(), graph, context)
        if match_value is not None:
            def _arrow1475(__unit: None=None) -> bool:
                fp: LDNode = match_value
                return LDFormalParameter.validate(fp, context)

            if _arrow1475():
                fp_1: LDNode = match_value
                return fp_1

            else: 
                raise Exception(("Property of `exampleOfWork` of object with @id `" + dt.Id) + "` was not a FormalParameter")


        else: 
            return None


    @staticmethod
    def set_example_of_work(pv: LDNode, example_of_work: LDNode, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDPropertyValue.example_of_work(), example_of_work, context)

    @staticmethod
    def validate(pv: LDNode, context: LDContext | None=None) -> bool:
        return pv.HasProperty(LDPropertyValue.name(), context) if pv.HasType(LDPropertyValue.schema_type(), context) else False

    @staticmethod
    def validate_component(pv: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1477:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1476(x: str, y: str) -> bool:
                    return x == y

                return _arrow1476

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return contains("Component", pv.AdditionalType, ObjectExpr1477()) if LDPropertyValue.validate(pv, context) else False

    @staticmethod
    def validate_parameter_value(pv: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1479:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1478(x: str, y: str) -> bool:
                    return x == y

                return _arrow1478

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        class ObjectExpr1481:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1480(x_1: str, y_1: str) -> bool:
                    return x_1 == y_1

                return _arrow1480

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return (True if contains("ParameterValue", pv.AdditionalType, ObjectExpr1479()) else contains("ProcessParameterValue", pv.AdditionalType, ObjectExpr1481())) if LDPropertyValue.validate(pv, context) else False

    @staticmethod
    def validate_characteristic_value(pv: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1483:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1482(x: str, y: str) -> bool:
                    return x == y

                return _arrow1482

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        class ObjectExpr1485:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1484(x_1: str, y_1: str) -> bool:
                    return x_1 == y_1

                return _arrow1484

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return (True if contains("CharacteristicValue", pv.AdditionalType, ObjectExpr1483()) else contains("MaterialAttributeValue", pv.AdditionalType, ObjectExpr1485())) if LDPropertyValue.validate(pv, context) else False

    @staticmethod
    def validate_factor_value(pv: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1487:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1486(x: str, y: str) -> bool:
                    return x == y

                return _arrow1486

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return contains("FactorValue", pv.AdditionalType, ObjectExpr1487()) if LDPropertyValue.validate(pv, context) else False

    @staticmethod
    def validate_fragment_descriptor(pv: LDNode, context: LDContext | None=None) -> bool:
        return (LDPropertyValue.get_name_as_string(pv, context) == "FragmentDescriptor") if LDPropertyValue.validate(pv, context) else False

    @staticmethod
    def validate_cwlparameter(pv: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1489:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1488(x: str, y: str) -> bool:
                    return x == y

                return _arrow1488

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return contains("WorkflowInput", pv.AdditionalType, ObjectExpr1489()) if (pv.HasProperty(LDPropertyValue.value(), context) if (pv.HasProperty(LDPropertyValue.example_of_work(), context) if LDPropertyValue.validate(pv, context) else False) else False) else False

    @staticmethod
    def validate_doi(pv: LDNode, context: LDContext | None=None) -> bool:
        if LDPropertyValue.validate(pv, context):
            match_value: str | None = LDPropertyValue.try_get_name_as_string(pv, context)
            match_value_1: str | None = LDPropertyValue.try_get_value_as_string(pv, context)
            match_value_2: str | None = LDPropertyValue.try_get_property_idas_string(pv, context)
            (pattern_matching_result,) = (None,)
            if match_value is not None:
                if match_value_1 is not None:
                    if match_value_2 is not None:
                        def _arrow1490(__unit: None=None) -> bool:
                            value: str = match_value_1
                            id: str = match_value_2
                            return (id == LDPropertyValue.doi_url()) if (match_value == LDPropertyValue.doi_key()) else False

                        if _arrow1490():
                            pattern_matching_result = 0

                        else: 
                            pattern_matching_result = 1


                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1

            if pattern_matching_result == 0:
                return True

            elif pattern_matching_result == 1:
                return False


        else: 
            return False


    @staticmethod
    def validate_pub_med_id(pv: LDNode, context: LDContext | None=None) -> bool:
        if LDPropertyValue.validate(pv, context):
            match_value: str | None = LDPropertyValue.try_get_name_as_string(pv, context)
            match_value_1: str | None = LDPropertyValue.try_get_value_as_string(pv, context)
            match_value_2: str | None = LDPropertyValue.try_get_property_idas_string(pv, context)
            (pattern_matching_result,) = (None,)
            if match_value is not None:
                if match_value_1 is not None:
                    if match_value_2 is not None:
                        def _arrow1491(__unit: None=None) -> bool:
                            value: str = match_value_1
                            id: str = match_value_2
                            return (id == LDPropertyValue.pubmed_idurl()) if (match_value == LDPropertyValue.pubmed_idkey()) else False

                        if _arrow1491():
                            pattern_matching_result = 0

                        else: 
                            pattern_matching_result = 1


                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1

            if pattern_matching_result == 0:
                return True

            elif pattern_matching_result == 1:
                return False


        else: 
            return False


    @staticmethod
    def validate_prefix(pv: LDNode, context: LDContext | None=None) -> bool:
        if LDPropertyValue.validate(pv, context):
            match_value: str | None = LDPropertyValue.try_get_name_as_string(pv, context)
            match_value_1: str | None = LDPropertyValue.try_get_value_as_string(pv, context)
            (pattern_matching_result,) = (None,)
            if match_value is not None:
                if match_value_1 is not None:
                    def _arrow1492(__unit: None=None) -> bool:
                        value: str = match_value_1
                        return match_value == LDPropertyValue.prefix_key()

                    if _arrow1492():
                        pattern_matching_result = 0

                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1

            if pattern_matching_result == 0:
                return True

            elif pattern_matching_result == 1:
                return False


        else: 
            return False


    @staticmethod
    def validate_position(pv: LDNode, context: LDContext | None=None) -> bool:
        if LDPropertyValue.validate(pv, context):
            match_value: str | None = LDPropertyValue.try_get_name_as_string(pv, context)
            match_value_1: str | None = LDPropertyValue.try_get_value_as_string(pv, context)
            (pattern_matching_result,) = (None,)
            if match_value is not None:
                if match_value_1 is not None:
                    def _arrow1493(__unit: None=None) -> bool:
                        value: str = match_value_1
                        return match_value == LDPropertyValue.position_key()

                    if _arrow1493():
                        pattern_matching_result = 0

                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1

            if pattern_matching_result == 0:
                return True

            elif pattern_matching_result == 1:
                return False


        else: 
            return False


    @staticmethod
    def validate_glob(pv: LDNode, context: LDContext | None=None) -> bool:
        if LDPropertyValue.validate(pv, context):
            match_value: str | None = LDPropertyValue.try_get_name_as_string(pv, context)
            match_value_1: str | None = LDPropertyValue.try_get_value_as_string(pv, context)
            (pattern_matching_result,) = (None,)
            if match_value is not None:
                if match_value_1 is not None:
                    def _arrow1494(__unit: None=None) -> bool:
                        value: str = match_value_1
                        return match_value == LDPropertyValue.glob_key()

                    if _arrow1494():
                        pattern_matching_result = 0

                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1

            if pattern_matching_result == 0:
                return True

            elif pattern_matching_result == 1:
                return False


        else: 
            return False


    @staticmethod
    def gen_id(name: str, value: str | None=None, property_id: str | None=None, prefix: str | None=None) -> str:
        prefix_1: str = default_arg(prefix, "PV")
        def _arrow1495(__unit: None=None) -> str:
            pid_1: str = property_id
            return ((("#" + prefix_1) + "_") + name) + ""

        def _arrow1496(__unit: None=None) -> str:
            value_2: str = value
            return ((((("#" + prefix_1) + "_") + name) + "_") + value_2) + ""

        def _arrow1497(__unit: None=None) -> str:
            pid: str = property_id
            value_1: str = value
            return ((((("#" + prefix_1) + "_") + name) + "_") + value_1) + ""

        return clean((_arrow1495() if (property_id is not None) else (((("#" + prefix_1) + "_") + name) + "")) if (value is None) else (_arrow1496() if (property_id is None) else _arrow1497()))

    @staticmethod
    def gen_id_component(name: str, value: str | None=None, property_id: str | None=None) -> str:
        return LDPropertyValue.gen_id(name, value, property_id, "Component")

    @staticmethod
    def gen_id_parameter_value(name: str, value: str | None=None, property_id: str | None=None) -> str:
        return LDPropertyValue.gen_id(name, value, property_id, "ParameterValue")

    @staticmethod
    def gen_id_characteristic_value(name: str, value: str | None=None, property_id: str | None=None) -> str:
        return LDPropertyValue.gen_id(name, value, property_id, "CharacteristicValue")

    @staticmethod
    def gen_id_factor_value(name: str, value: str | None=None, property_id: str | None=None) -> str:
        return LDPropertyValue.gen_id(name, value, property_id, "FactorValue")

    @staticmethod
    def gen_id_fragment_descriptor(file_name: str) -> str:
        return ("#Descriptor_" + file_name) + ""

    @staticmethod
    def gen_id_cwlparameter(name: str, values: Array[str]) -> str:
        def f(v: str) -> str:
            return replace(v, " ", "_")

        return ((("#WorkflowParameter_" + name) + "_") + (join("_", ResizeArray_map(f, values)) if (len(values) > 0) else "")) + ""

    @staticmethod
    def gen_id_prefix(prefix: str) -> str:
        return ("#Prefix_" + prefix) + ""

    @staticmethod
    def gen_id_position(position: int) -> str:
        return ("#Position_" + str(position)) + ""

    @staticmethod
    def gen_id_glob(glob: str) -> str:
        return ("#Glob_" + glob) + ""

    @staticmethod
    def create(name: str, value: str | None=None, id: str | None=None, property_id: str | None=None, unit_code: str | None=None, unit_text: str | None=None, value_reference: str | None=None, context: LDContext | None=None) -> LDNode:
        pv: LDNode = LDNode(LDPropertyValue.gen_id(name, value, property_id) if (id is None) else id, [LDPropertyValue.schema_type()], None, context)
        LDPropertyValue.set_name_as_string(pv, name, context)
        pv.SetOptionalProperty(LDPropertyValue.value(), value, context)
        def action(pid: str) -> None:
            LDPropertyValue.set_property_idas_string(pv, pid, context)

        iterate(action, to_array(property_id))
        def action_1(uc: str) -> None:
            LDPropertyValue.set_unit_code_as_string(pv, uc, context)

        iterate(action_1, to_array(unit_code))
        def action_2(ut: str) -> None:
            LDPropertyValue.set_unit_text_as_string(pv, ut, context)

        iterate(action_2, to_array(unit_text))
        def action_3(vr: str) -> None:
            LDPropertyValue.set_value_reference_as_string(pv, vr, context)

        iterate(action_3, to_array(value_reference))
        return pv

    @staticmethod
    def create_component(name: str, value: str | None=None, id: str | None=None, property_id: str | None=None, unit_code: str | None=None, unit_text: str | None=None, value_reference: str | None=None, context: LDContext | None=None) -> LDNode:
        id_1: str = LDPropertyValue.gen_id_component(name, value, property_id) if (id is None) else id
        c: LDNode = LDPropertyValue.create(name, value, id_1, property_id, unit_code, unit_text, value_reference, context)
        c.AdditionalType = ["Component"]
        return c

    @staticmethod
    def create_parameter_value(name: str, value: str | None=None, id: str | None=None, property_id: str | None=None, unit_code: str | None=None, unit_text: str | None=None, value_reference: str | None=None, context: LDContext | None=None) -> LDNode:
        id_1: str = LDPropertyValue.gen_id_parameter_value(name, value, property_id) if (id is None) else id
        pv: LDNode = LDPropertyValue.create(name, value, id_1, property_id, unit_code, unit_text, value_reference, context)
        pv.AdditionalType = ["ParameterValue"]
        return pv

    @staticmethod
    def create_characteristic_value(name: str, value: str | None=None, id: str | None=None, property_id: str | None=None, unit_code: str | None=None, unit_text: str | None=None, value_reference: str | None=None, context: LDContext | None=None) -> LDNode:
        id_1: str = LDPropertyValue.gen_id_characteristic_value(name, value, property_id) if (id is None) else id
        cv: LDNode = LDPropertyValue.create(name, value, id_1, property_id, unit_code, unit_text, value_reference, context)
        cv.AdditionalType = ["CharacteristicValue"]
        return cv

    @staticmethod
    def create_factor_value(name: str, value: str | None=None, id: str | None=None, property_id: str | None=None, unit_code: str | None=None, unit_text: str | None=None, value_reference: str | None=None, context: LDContext | None=None) -> LDNode:
        id_1: str = LDPropertyValue.gen_id_factor_value(name, value, property_id) if (id is None) else id
        fv: LDNode = LDPropertyValue.create(name, value, id_1, property_id, unit_code, unit_text, value_reference, context)
        fv.AdditionalType = ["FactorValue"]
        return fv

    @staticmethod
    def create_fragment_descriptor(file_name: str, value: str | None=None, id: str | None=None, unit_code: str | None=None, unit_text: str | None=None, value_reference: str | None=None, measurement_method: str | None=None, description: str | None=None, alternate_name: str | None=None, disambiguating_descriptions: Any | None=None, subject_of: Any | None=None, context: LDContext | None=None) -> LDNode:
        id_1: str = LDPropertyValue.gen_id_fragment_descriptor(file_name) if (id is None) else id
        fd: LDNode = LDPropertyValue.create("FragmentDescriptor", value, id_1, "https://github.com/nfdi4plants/ARC-specification/blob/dev/ISA-XLSX.md#datamap-table-sheets", unit_code, unit_text, value_reference, context)
        if measurement_method is not None:
            LDPropertyValue.set_measurement_method_as_string(fd, value_3(measurement_method), context)

        if description is not None:
            LDPropertyValue.set_description_as_string(fd, value_3(description), context)

        if alternate_name is not None:
            LDPropertyValue.set_alternate_name_as_string(fd, value_3(alternate_name), context)

        if disambiguating_descriptions is not None:
            LDPropertyValue.set_disambiguating_descriptions_as_string(fd, value_3(disambiguating_descriptions), context)

        if subject_of is not None:
            LDPropertyValue.set_subject_of(fd, value_3(subject_of), context)

        return fd

    @staticmethod
    def create_cwlparameter(example_of_work: LDNode, name: str, values: Array[str], context: LDContext | None=None) -> LDNode:
        id: str = LDPropertyValue.gen_id_cwlparameter(name, values)
        pv: LDNode = LDPropertyValue.create(name, None, id, None, None, None, None, context)
        if len(values) > 0:
            pv.SetProperty(LDPropertyValue.value(), values, context)

        pv.SetProperty(LDPropertyValue.example_of_work(), example_of_work, context)
        pv.AdditionalType = ["WorkflowInput"]
        return pv

    @staticmethod
    def create_doi(value: str, context: LDContext | None=None) -> LDNode:
        return LDPropertyValue.create(LDPropertyValue.doi_key(), value, value, LDPropertyValue.doi_url(), None, None, None, context)

    @staticmethod
    def create_pub_med_id(value: str, context: LDContext | None=None) -> LDNode:
        return LDPropertyValue.create(LDPropertyValue.pubmed_idkey(), value, value, LDPropertyValue.pubmed_idurl(), None, None, None, context)

    @staticmethod
    def create_prefix(value: str, context: LDContext | None=None) -> LDNode:
        id: str = LDPropertyValue.gen_id_prefix(value)
        return LDPropertyValue.create(LDPropertyValue.prefix_key(), value, id, None, None, None, None, context)

    @staticmethod
    def create_position(value: int, context: LDContext | None=None) -> LDNode:
        id: str = LDPropertyValue.gen_id_position(value)
        return LDPropertyValue.create(LDPropertyValue.position_key(), int32_to_string(value), id, None, None, None, None, context)

    @staticmethod
    def create_glob(value: str, context: LDContext | None=None) -> LDNode:
        id: str = LDPropertyValue.gen_id_glob(value)
        return LDPropertyValue.create(LDPropertyValue.glob_key(), value, id, None, None, None, None, context)

    @staticmethod
    def try_get_as_doi(pv: LDNode, context: LDContext | None=None) -> str | None:
        return LDPropertyValue.get_value_as_string(pv, context) if LDPropertyValue.validate_doi(pv, context) else None

    @staticmethod
    def try_get_as_pub_med_id(pv: LDNode, context: LDContext | None=None) -> str | None:
        return LDPropertyValue.get_value_as_string(pv, context) if LDPropertyValue.validate_pub_med_id(pv, context) else None

    @staticmethod
    def try_get_as_prefix(pv: LDNode, context: LDContext | None=None) -> str | None:
        return LDPropertyValue.get_value_as_string(pv, context) if LDPropertyValue.validate_prefix(pv, context) else None

    @staticmethod
    def try_get_as_position(pv: LDNode, context: LDContext | None=None) -> int | None:
        if LDPropertyValue.validate_position(pv, context):
            match_value: tuple[bool, int]
            out_arg: int = 0
            def _arrow1498(__unit: None=None) -> int:
                return out_arg

            def _arrow1499(v: int) -> None:
                nonlocal out_arg
                out_arg = v or 0

            match_value = (try_parse(LDPropertyValue.get_value_as_string(pv, context), 511, False, 32, FSharpRef(_arrow1498, _arrow1499)), out_arg)
            return match_value[1] if match_value[0] else None

        else: 
            return None


    @staticmethod
    def try_get_as_glob(pv: LDNode, context: LDContext | None=None) -> str | None:
        return LDPropertyValue.get_value_as_string(pv, context) if LDPropertyValue.validate_glob(pv, context) else None


LDPropertyValue_reflection = _expr1500

__all__ = ["LDPropertyValue_reflection"]




from __future__ import annotations
from dataclasses import dataclass
from ..dynamic_obj.dynamic_obj import (DynamicObj, DynamicObj_reflection)
from ..fable_library.reflection import (TypeInfo, string_type, option_type, record_type, array_type, class_type)
from ..fable_library.types import (Record, Array)
from .requirements import Requirement

def _expr503() -> TypeInfo:
    return record_type("ARCtrl.CWL.StepInput", [], StepInput, lambda: [("Id", string_type), ("Source", option_type(string_type)), ("DefaultValue", option_type(string_type)), ("ValueFrom", option_type(string_type))])


@dataclass(eq = False, repr = False, slots = True)
class StepInput(Record):
    Id: str
    Source: str | None
    DefaultValue: str | None
    ValueFrom: str | None
    @staticmethod
    def create(id: str, source: str | None=None, default_value: str | None=None, value_from: str | None=None) -> StepInput:
        return StepInput(id, source, default_value, value_from)


StepInput_reflection = _expr503

def _expr504() -> TypeInfo:
    return record_type("ARCtrl.CWL.StepOutput", [], StepOutput, lambda: [("Id", array_type(string_type))])


@dataclass(eq = False, repr = False, slots = True)
class StepOutput(Record):
    Id: Array[str]
    @staticmethod
    def create(id: Array[str]) -> StepOutput:
        return StepOutput(id)


StepOutput_reflection = _expr504

def _expr505() -> TypeInfo:
    return class_type("ARCtrl.CWL.WorkflowStep", None, WorkflowStep, DynamicObj_reflection())


class WorkflowStep(DynamicObj):
    def __init__(self, id: str, in_: Array[StepInput], out_: StepOutput, run: str, requirements: Array[Requirement] | None=None, hints: Array[Requirement] | None=None) -> None:
        super().__init__()
        self._id: str = id
        self._in: Array[StepInput] = in_
        self._out: StepOutput = out_
        self._run: str = run
        self._requirements: Array[Requirement] | None = requirements
        self._hints: Array[Requirement] | None = hints

    @property
    def Id(self, __unit: None=None) -> str:
        this: WorkflowStep = self
        return this._id

    @Id.setter
    def Id(self, id: str) -> None:
        this: WorkflowStep = self
        this._id = id

    @property
    def In(self, __unit: None=None) -> Array[StepInput]:
        this: WorkflowStep = self
        return this._in

    @In.setter
    def In(self, in_: Array[StepInput]) -> None:
        this: WorkflowStep = self
        this._in = in_

    @property
    def Out(self, __unit: None=None) -> StepOutput:
        this: WorkflowStep = self
        return this._out

    @Out.setter
    def Out(self, out_: StepOutput) -> None:
        this: WorkflowStep = self
        this._out = out_

    @property
    def Run(self, __unit: None=None) -> str:
        this: WorkflowStep = self
        return this._run

    @Run.setter
    def Run(self, run: str) -> None:
        this: WorkflowStep = self
        this._run = run

    @property
    def Requirements(self, __unit: None=None) -> Array[Requirement] | None:
        this: WorkflowStep = self
        return this._requirements

    @Requirements.setter
    def Requirements(self, requirements: Array[Requirement] | None=None) -> None:
        this: WorkflowStep = self
        this._requirements = requirements

    @property
    def Hints(self, __unit: None=None) -> Array[Requirement] | None:
        this: WorkflowStep = self
        return this._hints

    @Hints.setter
    def Hints(self, hints: Array[Requirement] | None=None) -> None:
        this: WorkflowStep = self
        this._hints = hints


WorkflowStep_reflection = _expr505

def WorkflowStep__ctor_4159DCB0(id: str, in_: Array[StepInput], out_: StepOutput, run: str, requirements: Array[Requirement] | None=None, hints: Array[Requirement] | None=None) -> WorkflowStep:
    return WorkflowStep(id, in_, out_, run, requirements, hints)


__all__ = ["StepInput_reflection", "StepOutput_reflection", "WorkflowStep_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Process.material_attribute import MaterialAttribute
from ...arctrl_core.Process.material_attribute_value import (MaterialAttributeValue, MaterialAttributeValue_createAsPV)
from ...arctrl_core.value import Value as Value_1
from ...fable_library.list import (choose, of_array, FSharpList)
from ...fable_library.option import map
from ...fable_library.seq import map as map_1
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (object, IOptionalGetter, IGetters)
from ...thoth_json_core.types import (IEncodable, Decoder_1, IEncoderHelpers_1)
from ..decode import Decode_uri
from ..encode import try_include
from ..idtable import encode
from ..ontology_annotation import (OntologyAnnotation_ISAJson_encoder, OntologyAnnotation_ISAJson_decoder)
from ..property_value import (encoder, decoder, gen_id)
from .material_attribute import (encoder as encoder_1, decoder as decoder_1)
from .value import (encoder as encoder_2, decoder as decoder_2)

__A_ = TypeVar("__A_")

ROCrate_encoder: Callable[[MaterialAttributeValue], IEncodable] = encoder

ROCrate_decoder: Decoder_1[MaterialAttributeValue] = decoder(MaterialAttributeValue_createAsPV)

def ISAJson_genID(oa: MaterialAttributeValue) -> str:
    return gen_id(oa)


def ISAJson_encoder(id_map: Any | None, oa: MaterialAttributeValue) -> IEncodable:
    def f(oa_1: MaterialAttributeValue, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        def chooser(tupled_arg: tuple[str, IEncodable | None], oa_1: Any=oa_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping, tupled_arg[1])

        def _arrow2482(value: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2481(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr2481()

        def _arrow2483(value_2: MaterialAttribute, oa_1: Any=oa_1) -> IEncodable:
            return encoder_1(id_map, value_2)

        def _arrow2484(value_3: Value_1, oa_1: Any=oa_1) -> IEncodable:
            return encoder_2(id_map, value_3)

        def _arrow2485(oa_3: OntologyAnnotation, oa_1: Any=oa_1) -> IEncodable:
            return OntologyAnnotation_ISAJson_encoder(id_map, oa_3)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2482, ISAJson_genID(oa_1)), try_include("category", _arrow2483, oa_1.Category), try_include("value", _arrow2484, oa_1.Value), try_include("unit", _arrow2485, oa_1.Unit)]))
        class ObjectExpr2486(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], oa_1: Any=oa_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_1))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
                return helpers_1.encode_object(arg)

        return ObjectExpr2486()

    if id_map is not None:
        def _arrow2487(oa_4: MaterialAttributeValue, id_map: Any=id_map, oa: Any=oa) -> str:
            return ISAJson_genID(oa_4)

        return encode(_arrow2487, f, oa, id_map)

    else: 
        return f(oa)



def _arrow2492(get: IGetters) -> MaterialAttributeValue:
    def _arrow2488(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("@id", Decode_uri)

    def _arrow2489(__unit: None=None) -> MaterialAttribute | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("category", decoder_1)

    def _arrow2490(__unit: None=None) -> Value_1 | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("value", decoder_2)

    def _arrow2491(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("unit", OntologyAnnotation_ISAJson_decoder)

    return MaterialAttributeValue(_arrow2488(), _arrow2489(), _arrow2490(), _arrow2491())


ISAJson_decoder: Decoder_1[MaterialAttributeValue] = object(_arrow2492)

__all__ = ["ROCrate_encoder", "ROCrate_decoder", "ISAJson_genID", "ISAJson_encoder", "ISAJson_decoder"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ...fable_library.option import default_arg
from ...fable_library.range import range_big_int
from ...fable_library.seq import (to_list, delay, map, collect, singleton, append, is_empty, empty, iterate)
from ...fable_library.util import (IEnumerable_1, ignore)
from ...thoth_json_core.decode import (object, IRequiredGetter, string, seq as seq_1, IOptionalGetter, IGetters)
from ...thoth_json_core.encode import seq
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..Cells.fs_cell import FsCell
from ..fs_column import FsColumn
from ..fs_row import FsRow
from ..fs_worksheet import FsWorksheet
from ..Tables.fs_table import FsTable
from .column import (encode_no_numbers as encode_no_numbers_1, encode as encode_2, decode as decode_2)
from .row import (encode_no_numbers, encode, decode as decode_1)
from .table import (encode as encode_1, decode)

__A_ = TypeVar("__A_")

def encode_rows(no_numbering: bool, sheet: FsWorksheet) -> IEncodable:
    sheet.RescanRows()
    def _arrow367(__unit: None=None, no_numbering: Any=no_numbering, sheet: Any=sheet) -> IEnumerable_1[IEncodable]:
        def _arrow366(r: int) -> IEncodable:
            def _arrow365(__unit: None=None) -> IEnumerable_1[FsCell]:
                def _arrow364(c: int) -> IEnumerable_1[FsCell]:
                    match_value: FsCell | None = sheet.CellCollection.TryGetCell(r, c)
                    return singleton(FsCell("")) if (match_value is None) else singleton(match_value)

                return collect(_arrow364, range_big_int(1, 1, sheet.MaxColumnIndex))

            return encode_no_numbers(to_list(delay(_arrow365)))

        return map(_arrow366, range_big_int(1, 1, sheet.MaxRowIndex))

    def mapping(row_1: FsRow, no_numbering: Any=no_numbering, sheet: Any=sheet) -> IEncodable:
        return encode(row_1)

    j_rows: IEncodable = seq(to_list(delay(_arrow367))) if no_numbering else seq(map(mapping, sheet.Rows))
    def _arrow372(__unit: None=None, no_numbering: Any=no_numbering, sheet: Any=sheet) -> IEnumerable_1[tuple[str, IEncodable]]:
        def _arrow369(__unit: None=None) -> IEncodable:
            value: str = sheet.Name
            class ObjectExpr368(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr368()

        def _arrow371(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
            def _arrow370(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
                return singleton(("rows", j_rows))

            return append(singleton(("tables", seq(map(encode_1, sheet.Tables)))) if (not is_empty(sheet.Tables)) else empty(), delay(_arrow370))

        return append(singleton(("name", _arrow369())), delay(_arrow371))

    values_1: IEnumerable_1[tuple[str, IEncodable]] = to_list(delay(_arrow372))
    class ObjectExpr373(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], no_numbering: Any=no_numbering, sheet: Any=sheet) -> Any:
            def mapping_2(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_1))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping_2, values_1)
            return helpers_1.encode_object(arg)

    return ObjectExpr373()


def _arrow375(builder: IGetters) -> FsWorksheet:
    row_index: int = 0
    n: str
    object_arg: IRequiredGetter = builder.Required
    n = object_arg.Field("name", string)
    ts: IEnumerable_1[FsTable] | None
    arg_3: Decoder_1[IEnumerable_1[FsTable]] = seq_1(decode)
    object_arg_1: IOptionalGetter = builder.Optional
    ts = object_arg_1.Field("tables", arg_3)
    def _arrow374(__unit: None=None) -> IEnumerable_1[tuple[int | None, IEnumerable_1[FsCell]]] | None:
        arg_5: Decoder_1[IEnumerable_1[tuple[int | None, IEnumerable_1[FsCell]]]] = seq_1(decode_1)
        object_arg_2: IOptionalGetter = builder.Optional
        return object_arg_2.Field("rows", arg_5)

    rs: IEnumerable_1[tuple[int | None, IEnumerable_1[FsCell]]] = default_arg(_arrow374(), empty())
    sheet: FsWorksheet = FsWorksheet(n)
    def action_1(tupled_arg: tuple[int | None, IEnumerable_1[FsCell]]) -> None:
        nonlocal row_index
        row_i: int | None = tupled_arg[0]
        col_index: int = 0
        row_i_1: int = ((row_index + 1) if (row_i is None) else row_i) or 0
        row_index = row_i_1 or 0
        r: FsRow = sheet.Row(row_i_1)
        def action(cell: FsCell, tupled_arg: Any=tupled_arg) -> None:
            nonlocal col_index
            col_i: int
            match_value: int = cell.ColumnNumber or 0
            col_i = (col_index + 1) if (match_value == 0) else match_value
            col_index = col_i or 0
            c: FsCell = r.Item(col_i)
            c.Value = cell.Value
            c.DataType = cell.DataType

        iterate(action, tupled_arg[1])

    iterate(action_1, rs)
    if ts is None:
        pass

    else: 
        def action_2(t: FsTable) -> None:
            ignore(sheet.AddTable(t))

        iterate(action_2, ts)

    return sheet


decode_rows: Decoder_1[FsWorksheet] = object(_arrow375)

def encode_columns(no_numbering: bool, sheet: FsWorksheet) -> IEncodable:
    sheet.RescanRows()
    def _arrow379(__unit: None=None, no_numbering: Any=no_numbering, sheet: Any=sheet) -> IEnumerable_1[IEncodable]:
        def _arrow378(c: int) -> IEncodable:
            def _arrow377(__unit: None=None) -> IEnumerable_1[FsCell]:
                def _arrow376(r: int) -> IEnumerable_1[FsCell]:
                    match_value: FsCell | None = sheet.CellCollection.TryGetCell(r, c)
                    return singleton(FsCell("")) if (match_value is None) else singleton(match_value)

                return collect(_arrow376, range_big_int(1, 1, sheet.MaxRowIndex))

            return encode_no_numbers_1(to_list(delay(_arrow377)))

        return map(_arrow378, range_big_int(1, 1, sheet.MaxColumnIndex))

    def mapping(col_1: FsColumn, no_numbering: Any=no_numbering, sheet: Any=sheet) -> IEncodable:
        return encode_2(col_1)

    j_columns: IEncodable = seq(to_list(delay(_arrow379))) if no_numbering else seq(map(mapping, sheet.Columns))
    def _arrow384(__unit: None=None, no_numbering: Any=no_numbering, sheet: Any=sheet) -> IEnumerable_1[tuple[str, IEncodable]]:
        def _arrow381(__unit: None=None) -> IEncodable:
            value: str = sheet.Name
            class ObjectExpr380(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr380()

        def _arrow383(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
            def _arrow382(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
                return singleton(("columns", j_columns))

            return append(singleton(("tables", seq(map(encode_1, sheet.Tables)))) if (not is_empty(sheet.Tables)) else empty(), delay(_arrow382))

        return append(singleton(("name", _arrow381())), delay(_arrow383))

    values_1: IEnumerable_1[tuple[str, IEncodable]] = to_list(delay(_arrow384))
    class ObjectExpr385(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], no_numbering: Any=no_numbering, sheet: Any=sheet) -> Any:
            def mapping_2(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_1))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping_2, values_1)
            return helpers_1.encode_object(arg)

    return ObjectExpr385()


def _arrow388(builder: IGetters) -> FsWorksheet:
    col_index: int = 0
    n: str
    object_arg: IRequiredGetter = builder.Required
    n = object_arg.Field("name", string)
    ts: IEnumerable_1[FsTable] | None
    arg_3: Decoder_1[IEnumerable_1[FsTable]] = seq_1(decode)
    object_arg_1: IOptionalGetter = builder.Optional
    ts = object_arg_1.Field("tables", arg_3)
    cs: IEnumerable_1[tuple[int | None, IEnumerable_1[FsCell]]]
    arg_5: Decoder_1[IEnumerable_1[tuple[int | None, IEnumerable_1[FsCell]]]] = seq_1(decode_2)
    object_arg_2: IRequiredGetter = builder.Required
    cs = object_arg_2.Field("columns", arg_5)
    sheet: FsWorksheet = FsWorksheet(n)
    def action_1(tupled_arg: tuple[int | None, IEnumerable_1[FsCell]]) -> None:
        nonlocal col_index
        col_i: int | None = tupled_arg[0]
        row_index: int = 0
        col_i_1: int = ((col_index + 1) if (col_i is None) else col_i) or 0
        col_index = col_i_1 or 0
        col: FsColumn = sheet.Column(col_i_1)
        def action(cell: FsCell, tupled_arg: Any=tupled_arg) -> None:
            nonlocal row_index
            row_i: int
            match_value: int = cell.RowNumber or 0
            row_i = (row_index + 1) if (match_value == 0) else match_value
            row_index = row_i or 0
            c: FsCell = col.Item(row_index)
            c.Value = cell.Value
            c.DataType = cell.DataType

        iterate(action, tupled_arg[1])

    iterate(action_1, cs)
    if ts is None:
        pass

    else: 
        def action_2(t: FsTable) -> None:
            ignore(sheet.AddTable(t))

        iterate(action_2, ts)

    return sheet


decode_columns: Decoder_1[FsWorksheet] = object(_arrow388)

__all__ = ["encode_rows", "decode_rows", "encode_columns", "decode_columns"]

